"""Action distribution for compatibility with RLlib's interface."""
from abc import ABCMeta, abstractmethod

from nnrl.nn.actor import DeterministicPolicy, StochasticPolicy
from ray.rllib.models.action_dist import ActionDistribution
from torch import nn


class IncompatibleDistClsError(Exception):
    """Exception raised for incompatible action distribution and NN module.

    Args:
        dist_cls: Action distribution class
        module: NN module
        err: AssertionError explaining the reason why distribution and
            module are incompatible

    Attributes:
        message: Human-readable text explaining what caused the incompatibility
    """

    def __init__(self, dist_cls: type, module: nn.Module, err: Exception):
        # pylint:disable=unused-argument
        msg = (
            f"Action distribution type {dist_cls} is incompatible"
            " with NN module of type {type(module)}. Reason:\n"
            "    {err}"
        )
        super().__init__(msg)
        self.message = msg


class BaseActionDist(ActionDistribution, metaclass=ABCMeta):
    """Base class for TorchPolicy action distributions."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.module = self.model.module

    @classmethod
    def check_model_compat(cls, model: nn.Module):
        """Assert the given NN module is compatible with the distribution.

        Raises:
            IncompatibleDistClsError: If `model` is incompatible with the
                distribution class
        """
        try:
            cls._check_model_compat(model)
        except AssertionError as err:
            raise IncompatibleDistClsError(cls, model, err) from err

    @classmethod
    @abstractmethod
    def _check_model_compat(cls, model: nn.Module):
        pass


class WrapStochasticPolicy(BaseActionDist):
    """Wraps an nn.Module with a stochastic actor and its inputs.

    Expects actor to be an instance of StochasticPolicy.
    """

    # pylint:disable=abstract-method
    valid_actor_cls = StochasticPolicy

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sampled_logp = None

    def sample(self):
        action, logp = self.module.actor.sample(**self.inputs)
        self._sampled_logp = logp
        return action, logp

    def deterministic_sample(self):
        return self.module.actor.deterministic(**self.inputs)

    def sampled_action_logp(self):
        return self._sampled_logp

    def logp(self, x):
        return self.module.actor.log_prob(value=x, **self.inputs)

    def entropy(self):
        return self.module.actor.entropy(**self.inputs)

    @classmethod
    def _check_model_compat(cls, model):
        assert hasattr(model, "actor"), f"NN model {type(model)} has no actor attribute"
        assert isinstance(model.actor, cls.valid_actor_cls), (
            f"Expected actor to be an instance of {cls.valid_actor_cls};"
            " found {type(model.actor)} instead."
        )


class WrapDeterministicPolicy(BaseActionDist):
    """Wraps an nn.Module with a deterministic actor and its inputs.

    Expects actor to be an instance of DeterministicPolicy.
    """

    # pylint:disable=abstract-method
    valid_actor_cls = valid_behavior_cls = DeterministicPolicy

    def sample(self):
        action = self.module.behavior(**self.inputs)
        return action, None

    def deterministic_sample(self):
        return self.module.actor(**self.inputs), None

    def sampled_action_logp(self):
        return None

    def logp(self, x):
        return None

    @classmethod
    def _check_model_compat(cls, model: nn.Module):
        assert hasattr(model, "actor"), f"NN model {type(model)} has no actor attribute"
        assert isinstance(model.actor, cls.valid_actor_cls), (
            f"Expected actor to be an instance of {cls.valid_actor_cls};"
            " found {type(model.actor)} instead."
        )

        assert hasattr(model, "behavior"), "NN has no behavior attribute"
        assert isinstance(model.actor, cls.valid_behavior_cls), (
            f"Expected behavior to be an instance of {cls.valid_behavior_cls};"
            " found {type(model.behavior)} instead."
        )



"""OpenAI Gym environments and utilities."""
from ray.tune import register_env

from .rewards import get_reward_fn, has_reward_fn
from .rewards import register as register_reward_fn
from .termination import get_termination_fn, has_termination_fn
from .termination import register as register_termination_fn
from .utils import get_env_creator, has_env_creator

__all__ = [
    "get_reward_fn",
    "has_reward_fn",
    "register_reward_fn",
    "get_termination_fn",
    "has_termination_fn",
    "register_termination_fn",
    "register_env",
    "get_env_creator",
    "has_env_creator",
]



"""Base implemenation for all NN policy modules."""
from typing import List

import numpy as np
from torch import nn


class Base(nn.Module):
    """Base class for all NN policy modules.

    Defines common boilerplate code and interfaces.
    """

    @staticmethod
    def get_initial_state() -> List[np.ndarray]:
        """Returns the initial state for recurrent modules.

        By default returns an empty list. Override this for actual recurrent modules.

        Warning:
            RLlib calls this method to initialize its networks. We don't need such
            initialization but implement this placeholder to avoid crashes.
        """
        return []



"""Wrappers and adapters for compatibility with RLlib's API."""
from typing import Dict, List

from gym import Space
from ray.rllib.models.torch.torch_modelv2 import TorchModelV2
from ray.rllib.utils.framework import TensorType
from ray.rllib.utils.typing import ModelConfigDict
from torch import nn


class WrapRawModule(TorchModelV2, nn.Module):
    """Wrapper around a basic PyTorch module.

    Mostly for compatibility with ray>=1.1.0.
    """

    def __init__(
        self,
        obs_space: Space,
        action_space: Space,
        module: nn.Module,
        num_outputs: int,
        model_config: ModelConfigDict,
    ):
        # pylint:disable=too-many-arguments
        TorchModelV2.__init__(
            self,
            obs_space,
            action_space,
            num_outputs,
            model_config,
            name="PyTorchModule",
        )
        nn.Module.__init__(self)

        self.module = module

    def forward(
        self,
        input_dict: Dict[str, TensorType],
        state: List[TensorType],
        seq_lens: TensorType,
    ) -> (TensorType, List[TensorType]):
        pass

    def value_function(self) -> TensorType:
        pass

    def import_from_h5(self, h5_file: str) -> None:
        pass



"""Losses for Importance Sampled Fitted V Iteration."""
from typing import Tuple

import torch
from nnrl.nn.actor import Alpha, StochasticPolicy
from nnrl.nn.critic import VValue
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from ray.rllib.utils import override
from torch import Tensor, nn

from raylab.utils.types import StatDict

from .abstract import Loss


class ISFittedVIteration(Loss):
    """Loss function for Importance Sampled Fitted V Iteration.

    Args:
        critic: state-value function
        target_critic: state-value function for the next state

    Attributes:
        gamma: discount factor
    """

    IS_RATIOS = "is_ratios"
    gamma: float = 0.99

    def __init__(self, critic: VValue, target_critic: VValue):
        self.critic = critic
        self.target_critic = target_critic
        self.batch_keys: Tuple[str, str] = (
            SampleBatch.CUR_OBS,
            self.IS_RATIOS,
            SampleBatch.NEXT_OBS,
            SampleBatch.REWARDS,
            SampleBatch.DONES,
        )

        self._loss_fn = nn.MSELoss(reduction="none")

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        """Compute loss for importance sampled fitted V iteration."""
        obs, is_ratios, next_obs, reward, done = self.unpack_batch(batch)

        with torch.no_grad():
            target = self.sampled_one_step_state_values(obs, next_obs, reward, done)

        pred = self.critic(obs)
        loss = torch.mean(is_ratios * self._loss_fn(pred, target) / 2)
        return loss, {"loss(critic)": loss.item()}

    def sampled_one_step_state_values(
        self, obs: Tensor, next_obs: Tensor, reward: Tensor, done: Tensor
    ) -> Tensor:
        """Bootstrapped approximation of true state-value using sampled transition."""
        del obs
        return torch.where(
            done, reward, reward + self.gamma * self.target_critic(next_obs)
        )


class ISSoftVIteration(ISFittedVIteration):
    """Loss function for Importance Sampled Soft V Iteration.

    Args:
        critic: state-value function
        target_critic: state-value function for the next state
        actor: stochastic policy

    Attributes:
        gamma: discount factor
    """

    # pylint:disable=too-few-public-methods
    gamma: float = 0.99

    def __init__(
        self,
        critic: VValue,
        target_critic: VValue,
        actor: StochasticPolicy,
        alpha: Alpha,
    ):
        self.actor = actor
        self.alpha = alpha
        super().__init__(critic, target_critic)

    @override(ISFittedVIteration)
    def sampled_one_step_state_values(
        self, obs: Tensor, next_obs: Tensor, reward: Tensor, done: Tensor
    ) -> Tensor:
        """Bootstrapped approximation of true state-value using sampled transition."""
        _, logp = self.actor.sample(obs)
        entropy = -logp

        reward = reward + self.alpha() * entropy
        return torch.where(
            done, reward, reward + self.gamma * self.target_critic(next_obs)
        )



"""Collection of type annotations."""
from typing import Callable, Dict, Tuple, Union

from torch import Tensor

DynamicsFn = Callable[[Tensor, Tensor], Tuple[Tensor, Tensor]]

RewardFn = Callable[[Tensor, Tensor, Tensor], Tensor]

StatDict = Dict[str, Union[float, int]]

TerminationFn = Callable[[Tensor, Tensor, Tensor], Tensor]



# pylint:disable=missing-module-docstring
import gym
import numpy as np
from gym.spaces import Box

from .mixins import IrrelevantRedundantMixin, RNGMixin
from .utils import assert_flat_box_space


class CorrelatedIrrelevant(IrrelevantRedundantMixin, RNGMixin, gym.ObservationWrapper):
    """Add correlated random variables to the environment's observations.

    Samples :math:`Uniform(0, 1)` random variables on reset and appends them to
    the observation. For every subsequent timestep t, appends the initial
    variables, exponentiated to the power of t, to the observation.

    Args:
        env: Gym environment
        size: Number of random reward-irrelevant variables
    """

    def __init__(self, env: gym.Env, size: int):
        assert_flat_box_space(env.observation_space, self)
        super().__init__(env)
        self._size = size
        self._timestep: int = 0
        self._uvars: np.ndarray = None

        original = self.observation_space
        low = np.concatenate([original.low, [0] * self._size])
        high = np.concatenate([original.high, [1] * self._size])
        self.observation_space = Box(
            low=low.astype(original.dtype), high=high.astype(original.dtype)
        )

        self._set_reward_if_possible()
        self._set_termination_if_possible()

    @property
    def added_size(self):
        return self._size

    def reset(self, **kwargs):
        self._uvars = self.np_random.uniform(size=self._size)
        self._timestep = 0
        return super().reset(**kwargs)

    def step(self, action):
        self._timestep += 1
        return super().step(action)

    def _added_vars(self, _) -> np.ndarray:
        return self._uvars ** self._timestep



"""Soft Actor-Critic.

Off-Policy Maximum Entropy Deep Reinforcement Learning with a Stochastic Actor.
"""
from raylab.agents import Trainer
from raylab.agents.off_policy import OffPolicyMixin
from raylab.options import configure

from .policy import SACTorchPolicy


@configure
@OffPolicyMixin.add_options
class SACTrainer(OffPolicyMixin, Trainer):
    """Single agent trainer for SAC."""

    _name = "SoftAC"
    _policy_class = SACTorchPolicy



"""SVG(inf) policy class using PyTorch."""
import torch
from nnrl.optim import build_optimizer
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from ray.rllib.utils import override
from torch import nn

from raylab.agents.svg import SVGTorchPolicy
from raylab.options import configure, option
from raylab.policy import AdaptiveKLCoeffMixin, EnvFnMixin, learner_stats
from raylab.policy.losses import TrajectorySVG
from raylab.policy.off_policy import OffPolicyMixin, off_policy_options
from raylab.utils.replay_buffer import ReplayField


@configure
@off_policy_options
@option(
    "updates_per_step",
    1.0,
    help="Model and Value function updates per step in the environment",
)
@option("max_grad_norm", 10.0, help="Clip gradient norms by this value")
@option("optimizer/on_policy", {"type": "Adam", "lr": 1e-3})
@option("optimizer/off_policy", {"type": "Adam", "lr": 1e-3})
@option(
    "kl_schedule/",
    help="Options for adaptive KL coefficient. See raylab.utils.adaptive_kl",
    allow_unknown_subkeys=True,
)
@option("module/type", default="SVG")
@option("exploration_config/type", "raylab.utils.exploration.StochasticActor")
class SVGInfTorchPolicy(OffPolicyMixin, AdaptiveKLCoeffMixin, SVGTorchPolicy):
    """Stochastic Value Gradients policy for full trajectories."""

    # pylint:disable=abstract-method,too-many-ancestors

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)
        self.loss_actor = TrajectorySVG(
            self.module.model,
            self.module.actor,
            self.module.critic,
        )

        self.build_replay_buffer()

    @override(OffPolicyMixin)
    def build_replay_buffer(self):
        super().build_replay_buffer()
        self.replay.add_fields(ReplayField(SampleBatch.ACTION_LOGP))

    @override(EnvFnMixin)
    def _set_reward_hook(self):
        self.loss_actor.set_reward_fn(self.reward_fn)

    @override(SVGTorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["optimizer"]
        component_map = {
            "on_policy": self.module.actor,
            "off_policy": nn.ModuleList([self.module.model, self.module.critic]),
        }

        mapping = {
            name: build_optimizer(module, config[name])
            for name, module in component_map.items()
        }

        optimizers.update(mapping)
        return optimizers

    @override(SVGTorchPolicy)
    def compile(self):
        super().compile()
        self.loss_actor.compile()

    @override(OffPolicyMixin)
    def improve_policy(self, _):
        pass

    @learner_stats
    @override(SVGTorchPolicy)
    def learn_on_batch(self, samples: SampleBatch) -> dict:
        traj_len = samples.count
        self.add_to_buffer(samples)

        info = {}
        for _ in range(int(traj_len * self.config["updates_per_step"])):
            batch = self.replay.sample(self.config["batch_size"])
            batch = self.lazy_tensor_dict(batch)
            off_policy_stats = self._learn_off_policy(batch)

        info.update(off_policy_stats)
        info.update(self._learn_on_policy(samples))
        return info

    def _learn_off_policy(self, batch: TensorDict) -> dict:
        """Update off-policy components."""
        batch, info = self.add_truncated_importance_sampling_ratios(batch)

        with self.optimizers.optimize("off_policy"):
            loss, _info = self.compute_joint_model_value_loss(batch)
            info.update(_info)
            loss.backward()

        info.update(self.extra_grad_info(batch, on_policy=False))
        self._update_polyak()
        return info

    def _learn_on_policy(self, samples: SampleBatch) -> dict:
        """Update on-policy components."""
        batch = self.lazy_tensor_dict(samples)
        episodes = [self.lazy_tensor_dict(s) for s in samples.split_by_episode()]

        with self.optimizers.optimize("on_policy"):
            loss, info = self.loss_actor(episodes)
            kl_div = self._avg_kl_divergence(batch)
            loss = loss + kl_div * self.curr_kl_coeff
            loss.backward()

        info.update(self.extra_grad_info(batch, on_policy=True))
        info.update(self.update_kl_coeff(samples))
        return info

    @torch.no_grad()
    @override(AdaptiveKLCoeffMixin)
    def _kl_divergence(self, sample_batch: SampleBatch):
        batch_tensors = self.lazy_tensor_dict(sample_batch)
        return self._avg_kl_divergence(batch_tensors).item()

    def _avg_kl_divergence(self, batch_tensors: TensorDict):
        logp = self.module.actor.log_prob(
            batch_tensors[SampleBatch.CUR_OBS], batch_tensors[SampleBatch.ACTIONS]
        )
        return torch.mean(batch_tensors[SampleBatch.ACTION_LOGP] - logp)

    @torch.no_grad()
    def extra_grad_info(self, batch_tensors: TensorDict, on_policy: bool) -> dict:
        """Compute gradient norm for components. Also clips on-policy gradient."""
        if on_policy:
            params = self.module.actor.parameters()
            max_norm = self.config["max_grad_norm"]
            fetches = {
                "policy_grad_norm": nn.utils.clip_grad_norm_(params, max_norm),
                "policy_entropy": -batch_tensors[SampleBatch.ACTION_LOGP].mean(),
                "curr_kl_coeff": self.curr_kl_coeff,
            }
        else:
            max_norm = float("inf")
            fetches = {}
            params = self.module.model.parameters()
            fetches["model_grad_norm"] = nn.utils.clip_grad_norm_(params, max_norm)
            params = self.module.critic.parameters()
            fetches["value_grad_norm"] = nn.utils.clip_grad_norm_(params, max_norm)

        return {k: (v.item() if torch.is_tensor(v) else v) for k, v in fetches.items()}



"""Registry of environment reward functions in PyTorch."""
import math
from typing import Callable, Optional

import torch
from torch import nn

from raylab.tune.registry import RAYLAB_REWARD, _raylab_registry

from .utils import get_env_parameters, has_env_creator

# For testing purposes
REWARDS = {}


def register(*ids: str) -> Callable[[type], type]:
    """Register reward function class for environments with given ids."""

    def librarian(cls):
        for id_ in ids:
            REWARDS[id_] = cls
            _raylab_registry.register(RAYLAB_REWARD, id_, cls)

        return cls

    return librarian


def has_reward_fn(env_id: str) -> bool:
    """Whether the environment id has a reward function in the global registry."""
    return _raylab_registry.contains(RAYLAB_REWARD, env_id)


def get_reward_fn(env_id: str, env_config: Optional[dict] = None) -> "RewardFn":
    """Return the reward funtion for the given environment name and configuration.

    Only returns reward functions for environments which have been registered with Tune.
    """
    assert has_env_creator(env_id), f"{env_id} environment not registered with Tune."
    assert has_reward_fn(env_id), f"{env_id} environment reward not registered."

    env_config = env_config or {}
    reward_fn = _raylab_registry.get(RAYLAB_REWARD, env_id)(env_config)
    if env_config.get("time_aware", False):
        reward_fn = TimeAwareRewardFn(reward_fn)
    return reward_fn


class RewardFn(nn.Module):
    """Module that computes an environment's reward function for batches of inputs."""

    def __init__(self, _):
        super().__init__()

    def forward(self, state, action, next_state):  # pylint:disable=arguments-differ
        raise NotImplementedError


class TimeAwareRewardFn(RewardFn):
    """Wraps a reward function and removes time dimension before forwarding."""

    def __init__(self, reward_fn):
        super().__init__(None)
        self.reward_fn = reward_fn

    def forward(self, state, action, next_state):
        return self.reward_fn(state[..., :-1], action, next_state[..., :-1])


################################################################################
# Built-ins
################################################################################


@register("CartPoleSwingUp-v0", "TorchCartPoleSwingUp-v0")
class CartPoleSwingUpV0Reward(RewardFn):
    """
    Compute CartPoleSwingUp's reward given a possibly batched transition.
    Assumes all but the last dimension are batch ones.
    """

    def forward(self, state, action, next_state):
        return next_state[..., 2]


@register("CartPoleSwingUp-v1", "TorchCartPoleSwingUp-v1")
class CartPoleSwingUpV1Reward(RewardFn):
    """
    Compute CartPoleSwingUp's reward given a possibly batched transition.
    Assumes all but the last dimension are batch ones.
    """

    def forward(self, state, action, next_state):
        return (1 + next_state[..., 2]) / 2


@register("HalfCheetah-v3")
class HalfCheetahReward(RewardFn):
    """Compute rewards given a possibly batched transition.

    Assumes all but the last dimension are batch ones.
    """

    def __init__(self, config):
        super().__init__(config)
        kwargs = config.get("kwargs", {})
        parameters = get_env_parameters("HalfCheetah-v3")
        for attr in """
        ctrl_cost_weight
        forward_reward_weight
        exclude_current_positions_from_observation
        """.split():
            setattr(self, "_" + attr, kwargs.get(attr, parameters[attr].default))

        assert (
            self._exclude_current_positions_from_observation is False
        ), "Need x position for HalfCheetah-v3 reward function"
        self.delta_t = 0.05

    def forward(self, state, action, next_state):
        x_position_before = state[..., 0]
        x_position_after = next_state[..., 0]
        x_velocity = (x_position_after - x_position_before) / self.delta_t

        control_cost = self._ctrl_cost_weight * (action ** 2).sum(dim=-1)

        forward_reward = self._forward_reward_weight * x_velocity

        return forward_reward - control_cost


@register("HVAC")
class HVACReward(RewardFn):
    """Compute HVAC's reward function."""

    def __init__(self, config):
        super().__init__(config)
        from .environments.hvac import DEFAULT_CONFIG

        config = {**DEFAULT_CONFIG, **config}
        self.air_max = torch.as_tensor(config["AIR_MAX"]).float()
        self.is_room = torch.as_tensor(config["IS_ROOM"])
        self.cost_air = torch.as_tensor(config["COST_AIR"]).float()
        self.temp_low = torch.as_tensor(config["TEMP_LOW"]).float()
        self.temp_up = torch.as_tensor(config["TEMP_UP"]).float()
        self.penalty = torch.as_tensor(config["PENALTY"]).float()

    def forward(self, state, action, next_state):
        air = action * self.air_max
        temp = next_state[..., :-1]

        reward = -(
            self.is_room
            * (
                air * self.cost_air
                + ((temp < self.temp_low) | (temp > self.temp_up)) * self.penalty
                + 10.0 * torch.abs((self.temp_up + self.temp_low) / 2.0 - temp)
            )
        ).sum(dim=-1)

        return reward


@register("IndustrialBenchmark-v0")
class IndustrialBenchmarkReward(RewardFn):
    """IndustrialBenchmarks's reward function."""

    def __init__(self, config):
        super().__init__(config)
        self.reward_type = config.get("reward_type", "classic")

    def forward(self, state, action, next_state):
        con_coeff, fat_coeff = 1, 3
        consumption, fatigue = next_state[..., 4], next_state[..., 5]
        reward = -(con_coeff * consumption + fat_coeff * fatigue)

        if self.reward_type == "delta":
            old_consumption, old_fatigue = state[..., 4], state[..., 5]
            old_reward = -(con_coeff * old_consumption + fat_coeff * old_fatigue)
            reward -= old_reward

        return reward / 100


@register("Navigation")
class NavigationReward(RewardFn):
    """Navigation's reward function."""

    def __init__(self, config):
        super().__init__(config)
        from .environments.navigation import DEFAULT_CONFIG

        config = {**DEFAULT_CONFIG, **config}
        self._end = torch.as_tensor(config["end"]).float()

    def forward(self, state, action, next_state):
        next_state = next_state[..., :2]
        goal = self._end
        return -torch.norm(next_state - goal, p=2, dim=-1)


@register("Reacher-v2")
class ReacherReward(RewardFn):
    """Reacher-v3's reward function."""

    def forward(self, state, action, next_state):
        dist = state[..., -3:]
        reward_dist = -torch.norm(dist, dim=-1)
        reward_ctrl = -torch.sum(action ** 2, dim=-1)
        return reward_dist + reward_ctrl


@register("Reservoir")
class ReservoirReward(RewardFn):
    """Reservoir's reward function."""

    def __init__(self, config):
        super().__init__(config)
        from .environments.reservoir import DEFAULT_CONFIG

        config = {**DEFAULT_CONFIG, **config}
        self.lower_bound = torch.as_tensor(config["LOWER_BOUND"])
        self.upper_bound = torch.as_tensor(config["UPPER_BOUND"])

        self.low_penalty = torch.as_tensor(config["LOW_PENALTY"])
        self.high_penalty = torch.as_tensor(config["HIGH_PENALTY"])

    def forward(self, state, action, next_state):
        rlevel = next_state[..., :-1]

        mean_capacity_deviation = -0.01 * torch.abs(
            rlevel - (self.lower_bound + self.upper_bound) / 2
        )
        overflow_penalty = self.high_penalty * torch.max(
            torch.zeros_like(rlevel), rlevel - self.upper_bound
        )
        underflow_penalty = self.low_penalty * torch.max(
            torch.zeros_like(rlevel), self.lower_bound - rlevel
        )

        penalty = mean_capacity_deviation + overflow_penalty + underflow_penalty
        return penalty.sum(dim=-1)


@register("MountainCarContinuous-v0")
class MountainCarContinuousReward(RewardFn):
    """MountainCarContinuous' reward function."""

    def __init__(self, config):
        super().__init__(config)
        goal_position = 0.45
        goal_velocity = config.get("goal_velocity", 0.0)
        self.goal = torch.as_tensor([goal_position, goal_velocity])

    def forward(self, state, action, next_state):
        done = (next_state >= self.goal).all(-1)
        shape = state.shape[:-1]
        reward = torch.where(done, torch.empty(shape).fill_(200), torch.zeros(shape))
        reward -= torch.pow(action, 2).squeeze(-1) * 0.1
        return reward


@register("ReacherBulletEnv-v0")
class ReacherBulletEnvReward(RewardFn):
    """ReacherBulletEnv-v0's reward function."""

    def forward(self, state, action, next_state):
        to_target_vec_old = state[..., 2:4]
        to_target_vec_new = next_state[..., 2:4]
        potential_old = -100 * to_target_vec_old.norm(p=2, dim=-1)
        potential_new = -100 * to_target_vec_new.norm(p=2, dim=-1)

        theta_dot = next_state[..., -3]
        gamma = next_state[..., -2]
        gamma_dot = next_state[..., -1]

        electricity_cost = -0.10 * (
            (action[..., 0] * theta_dot).abs() + (action[..., 1] * gamma_dot).abs()
        ) - 0.01 * (action[..., 0].abs() + action[..., 1].abs())

        stuck_joint_cost = torch.where(
            (gamma.abs() - 1).abs() < 0.01,
            torch.empty_like(electricity_cost).fill_(-0.1),
            torch.zeros_like(electricity_cost),
        )

        rewards = (potential_new - potential_old) + electricity_cost + stuck_joint_cost
        return rewards


@register("Pendulum-v0")
class PendulumReward(RewardFn):
    """Pendulum-v0's reward function."""

    def __init__(self, config):
        super().__init__(config)
        self.max_torque = 2.0

    def forward(self, state, action, next_state):
        # pylint:disable=invalid-name
        cos_th, sin_th, thdot = state[..., 0], state[..., 1], state[..., 2]
        th = torch.atan2(sin_th, cos_th)
        u = action[..., 0]

        # angle_normalize
        th = ((th + math.pi) % (2 * math.pi)) - math.pi

        u = torch.clamp(u, -self.max_torque, self.max_torque)
        costs = (th) ** 2 + 0.1 * thdot ** 2 + 0.001 * (u ** 2)
        return -costs


@register("Pusher-v2")
class PusherReward(RewardFn):
    """Pusher-v2's reward function."""

    vec_size: int = 3

    def forward(self, state, action, next_state):
        idx, vec_size = 14, self.vec_size
        tips_arm = state[..., idx : idx + vec_size]
        idx += vec_size
        obj = state[..., idx : idx + vec_size]
        idx += vec_size
        goal = state[..., idx : idx + vec_size]
        vec_1 = obj - tips_arm
        vec_2 = obj - goal

        reward_near = -torch.norm(vec_1, dim=-1)
        reward_dist = -torch.norm(vec_2, dim=-1)
        reward_ctrl = -torch.square(action).sum(dim=-1)
        return reward_dist + 0.1 * reward_ctrl + 0.5 * reward_near


@register("Walker2d-v3")
class Walker2DReward(RewardFn):
    """Walker2d-v3's reward function."""

    def __init__(self, config):
        super().__init__(config)
        kwargs = config.get("kwargs", {})
        parameters = get_env_parameters("Walker2d-v3")
        for attr in """
        ctrl_cost_weight
        forward_reward_weight
        healthy_reward
        terminate_when_unhealthy
        healthy_z_range
        healthy_angle_range
        exclude_current_positions_from_observation
        """.split():
            setattr(self, "_" + attr, kwargs.get(attr, parameters[attr].default))

        assert (
            not self._exclude_current_positions_from_observation
        ), "Need x position for Walkered-v3 reward function"
        self.delta_t = 0.008

    def forward(self, state, action, next_state):
        ctrl_cost = self._control_cost(action)

        x_velocity = (next_state[..., 0] - state[..., 0]) / self.delta_t
        forward_reward = self._forward_reward_weight * x_velocity

        healthy_reward = torch.full_like(ctrl_cost, fill_value=self._healthy_reward)
        if not self._terminate_when_unhealthy:
            healthy_reward = torch.where(
                self._is_healthy(next_state),
                healthy_reward,
                torch.zeros_like(healthy_reward),
            )

        return forward_reward + healthy_reward - ctrl_cost

    def _control_cost(self, action):
        control_cost = self._ctrl_cost_weight * torch.sum(torch.square(action), dim=-1)
        return control_cost

    def _is_healthy(self, state):
        # pylint:disable=invalid-name
        z, angle = state[..., 1], state[..., 2]

        min_z, max_z = self._healthy_z_range
        min_angle, max_angle = self._healthy_angle_range

        healthy_z = (min_z < z) & (z < max_z)
        healthy_angle = (min_angle < angle) & (angle < max_angle)
        is_healthy = healthy_z & healthy_angle

        return is_healthy


@register("Swimmer-v3")
class SwimmerReward(RewardFn):
    """Swimmer-v3's reward function."""

    def __init__(self, config):
        super().__init__(config)
        kwargs = config.get("kwargs", {})
        parameters = get_env_parameters("Swimmer-v3")
        for attr in """
        ctrl_cost_weight
        forward_reward_weight
        exclude_current_positions_from_observation
        """.split():
            setattr(self, "_" + attr, kwargs.get(attr, parameters[attr].default))

        assert (
            not self._exclude_current_positions_from_observation
        ), "Need x position for Swimmer-v3 reward function"
        self.delta_t = 0.04

    def forward(self, state, action, next_state):
        x_velocity = (next_state[..., 0] - state[..., 0]) / self.delta_t
        forward_reward = self._forward_reward_weight * x_velocity

        ctrl_cost = self._control_cost(action)

        return forward_reward - ctrl_cost

    def _control_cost(self, action):
        control_cost = self._ctrl_cost_weight * torch.sum(torch.square(action), dim=-1)
        return control_cost


@register("Hopper-v3")
class HopperReward(RewardFn):
    """Hopper-v3's reward function."""

    def __init__(self, config: dict):
        super().__init__(config)
        kwargs = config.get("kwargs", {})
        parameters = get_env_parameters("Hopper-v3")
        for attr in """
        ctrl_cost_weight
        exclude_current_positions_from_observation
        forward_reward_weight
        healthy_angle_range
        healthy_reward
        healthy_state_range
        healthy_z_range
        terminate_when_unhealthy
        """.split():
            setattr(self, "_" + attr, kwargs.get(attr, parameters[attr].default))

        assert (
            not self._exclude_current_positions_from_observation
        ), "Need current state position for Hopper-v3 reward function"

        self.delta_t = 0.008

    def forward(self, state, action, next_state):
        ctrl_cost = self.control_cost(action)

        x_velocity = (next_state[..., 0] - state[..., 0]) / self.delta_t
        forward_reward = self._forward_reward_weight * x_velocity

        healthy_reward = torch.full_like(ctrl_cost, fill_value=self._healthy_reward)
        if not self._terminate_when_unhealthy:
            healthy_reward = torch.where(
                self._is_healthy(next_state),
                healthy_reward,
                torch.zeros_like(healthy_reward),
            )

        return forward_reward + healthy_reward - ctrl_cost

    def control_cost(self, action):
        return self._ctrl_cost_weight * torch.sum(torch.square(action), dim=-1)

    def is_healthy(self, state):
        # pylint:disable=invalid-name
        z, angle, state_ = state[..., 1], state[..., 2], state[..., 2:]

        min_state, max_state = self._healthy_state_range
        min_z, max_z = self._healthy_z_range
        min_angle, max_angle = self._healthy_angle_range

        healthy_state = torch.all((min_state < state_) & (state_ < max_state), dim=-1)
        healthy_z = (min_z < z) & (z < max_z)
        healthy_angle = (min_angle < angle) & (angle < max_angle)

        return healthy_state & healthy_z & healthy_angle


@register("InvertedPendulum-v2")
class InvertedPendulumReward(RewardFn):
    # pylint:disable=abstract-method,missing-class-docstring
    def forward(self, state, action, next_state):
        # pylint:disable=no-self-use
        return torch.ones(next_state.shape[:-1])



# pylint:disable=missing-module-docstring
import functools
from typing import Any, Callable

from ray.rllib.policy.policy import LEARNER_STATS_KEY


def learner_stats(func: Callable[[Any], dict]) -> Callable[[Any], dict]:
    """Wrap function to return stats under learner stats key."""

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        stats = func(*args, **kwargs)
        nested = stats.get(LEARNER_STATS_KEY, {})
        unnested = {k: v for k, v in stats.items() if k != LEARNER_STATS_KEY}
        return {LEARNER_STATS_KEY: {**nested, **unnested}}

    return wrapped



"""SoftSVG policy class using PyTorch."""
import torch
from nnrl.optim import build_optimizer
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from ray.rllib.utils import override
from torch import nn

from raylab.agents.svg import SVGTorchPolicy
from raylab.options import configure, option
from raylab.policy import EnvFnMixin
from raylab.policy.losses import ISSoftVIteration, MaximumEntropyDual, OneStepSoftSVG
from raylab.policy.off_policy import OffPolicyMixin, off_policy_options
from raylab.utils.replay_buffer import ReplayField


def default_optimizer() -> dict:
    # pylint:disable=missing-function-docstring
    return {
        "model": {"type": "Adam", "lr": 1e-3},
        "actor": {"type": "Adam", "lr": 1e-3},
        "critic": {"type": "Adam", "lr": 1e-3},
        "alpha": {"type": "Adam", "lr": 1e-3},
    }


@configure
@off_policy_options
@option(
    "target_entropy",
    None,
    help="""Target entropy to optimize the temperature parameter towards

    If "auto", will use the heuristic provided in the SAC paper,
    H = -dim(A), where A is the action space
    """,
)
@option("optimizer", default_optimizer(), override=True)
@option("module/type", "SoftSVG")
@option("exploration_config/type", "raylab.utils.exploration.StochasticActor")
@option("exploration_config/pure_exploration_steps", 1000)
class SoftSVGTorchPolicy(OffPolicyMixin, SVGTorchPolicy):
    """Stochastic Value Gradients policy for off-policy learning."""

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)
        module = self.module
        self.loss_actor = OneStepSoftSVG(
            module.model, module.actor, module.critic, module.alpha
        )
        self.loss_actor.gamma = self.config["gamma"]

        self.loss_critic = ISSoftVIteration(
            module.critic, module.target_critic, module.actor, module.alpha
        )
        self.loss_critic.gamma = self.config["gamma"]

        target_entropy = (
            -action_space.shape[0]
            if self.config["target_entropy"] == "auto"
            else self.config["target_entropy"]
        )
        self.loss_alpha = MaximumEntropyDual(
            module.alpha, module.actor.sample, target_entropy
        )

        self.build_replay_buffer()

    def build_replay_buffer(self):
        super().build_replay_buffer()
        self.replay.add_fields(ReplayField(SampleBatch.ACTION_LOGP))

    @override(EnvFnMixin)
    def _set_reward_hook(self):
        self.loss_actor.set_reward_fn(self.reward_fn)

    @override(SVGTorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["optimizer"]
        module = self.module
        components = {
            "model": module.model,
            "actor": module.actor,
            "critic": module.critic,
            "alpha": module.alpha,
        }

        mapping = {
            name: build_optimizer(module, config[name])
            for name, module in components.items()
        }

        optimizers.update(mapping)
        return optimizers

    @torch.no_grad()
    @override(SVGTorchPolicy)
    def add_truncated_importance_sampling_ratios(self, batch_tensors):
        """Compute and add truncated importance sampling ratios to tensor batch."""
        batch = batch_tensors
        curr_logp = self.module.actor.log_prob(
            batch[SampleBatch.CUR_OBS], batch[SampleBatch.ACTIONS]
        )

        is_ratios = torch.exp(curr_logp - batch[SampleBatch.ACTION_LOGP])
        _is_ratios = torch.clamp(is_ratios, max=self.config["max_is_ratio"])

        batch[self.loss_actor.IS_RATIOS] = _is_ratios
        batch[self.loss_critic.IS_RATIOS] = _is_ratios

        info = {
            "is_ratio_max": is_ratios.max().item(),
            "is_ratio_mean": is_ratios.mean().item(),
            "is_ratio_min": is_ratios.min().item(),
            "cross_entropy": -curr_logp.mean().item(),
        }
        return batch, info

    @override(OffPolicyMixin)
    def improve_policy(self, batch: TensorDict) -> dict:
        batch, info = self.add_truncated_importance_sampling_ratios(batch)

        info.update(self._update_model(batch))
        info.update(self._update_critic(batch))
        info.update(self._update_actor(batch))
        if self.config["target_entropy"] is not None:
            info.update(self._update_alpha(batch))

        self._update_polyak()
        return info

    def _update_model(self, batch: TensorDict) -> dict:
        with self.optimizers.optimize("model"):
            model_loss, info = self.loss_model(batch)
            model_loss.backward()

        info.update(self.extra_grad_info("model"))
        return info

    def _update_critic(self, batch: TensorDict) -> dict:
        with self.optimizers.optimize("critic"):
            value_loss, info = self.loss_critic(batch)
            value_loss.backward()

        info.update(self.extra_grad_info("critic"))
        return info

    def _update_actor(self, batch: TensorDict) -> dict:
        with self.optimizers.optimize("actor"):
            svg_loss, info = self.loss_actor(batch)
            svg_loss.backward()

        info.update(self.extra_grad_info("actor"))
        return info

    def _update_alpha(self, batch: TensorDict) -> dict:
        with self.optimizers.optimize("alpha"):
            alpha_loss, info = self.loss_alpha(batch)
            alpha_loss.backward()

        info.update(self.extra_grad_info("alpha"))
        return info

    @torch.no_grad()
    def extra_grad_info(self, component: str) -> dict:
        """Return gradient statistics for component."""
        fetches = {
            f"grad_norm({component})": nn.utils.clip_grad_norm_(
                getattr(self.module, component).parameters(), float("inf")
            ).item()
        }
        return fetches



# pylint:disable=import-outside-toplevel
"""Registry of custom Gym environments."""
import importlib

import gym

from .utils import wrap_if_needed


def filtered_gym_env_ids():
    """
    Return environment ids in Gym registry for which all dependencies are installed.
    """
    specs = set(gym.envs.registry.all())

    if importlib.util.find_spec("atari_py") is None:
        specs.difference_update({s for s in specs if "atari" in s.entry_point})
    if importlib.util.find_spec("mujoco_py") is None:
        specs.difference_update({s for s in specs if "mujoco" in s.entry_point})
        specs.difference_update({s for s in specs if "robotics" in s.entry_point})
    if importlib.util.find_spec("Box2D") is None:
        specs.difference_update({s for s in specs if "box2d" in s.entry_point})

    return {s.id for s in specs}


IDS = filtered_gym_env_ids()
# kwarg trick from:
# https://github.com/satwikkansal/wtfpython#-the-sticky-output-function
ENVS = {
    i: wrap_if_needed(lambda config, i=i: gym.make(i, **config.get("kwargs", {})))
    for i in IDS
}


def register_external_library_environments(library_name):
    """Conveniency function for adding external environments to the global registry."""
    if importlib.util.find_spec(library_name) is None:
        return

    importlib.import_module(library_name)
    new_ids = filtered_gym_env_ids() - IDS
    for name in new_ids:

        @wrap_if_needed
        def _env_maker(config, env_id=name):
            importlib.import_module(library_name)

            kwargs = config.get("kwargs", {})
            return gym.make(env_id, **kwargs)

        ENVS[name] = _env_maker
    IDS.update(new_ids)


@wrap_if_needed
def _cartpole_stateless_maker(_):
    from raylab.envs.environments.cartpole_stateless import CartPoleStateless

    return CartPoleStateless()


@wrap_if_needed
def _navigation_maker(config):
    from raylab.envs.environments.navigation import NavigationEnv

    return NavigationEnv(config)


@wrap_if_needed
def _reservoir_maker(config):
    from raylab.envs.environments.reservoir import ReservoirEnv

    return ReservoirEnv(config)


@wrap_if_needed
def _hvac_maker(config):
    from raylab.envs.environments.hvac import HVACEnv

    return HVACEnv(config)


ENVS.update(
    {
        "CartPoleStateless": _cartpole_stateless_maker,
        "Navigation": _navigation_maker,
        "Reservoir": _reservoir_maker,
        "HVAC": _hvac_maker,
    }
)


register_external_library_environments("gym_cartpole_swingup")
register_external_library_environments("gym_industrial")
register_external_library_environments("pybullet_envs")



"""CLI for launching Tune experiments."""
import click

from .utils import tune_experiment


@tune_experiment
@click.argument("run_or_experiment", type=str)
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=False),
    default=None,
    help="Algorithm-specific configuration for Tune variant generation "
    "(e.g. env, hyperparams). Defaults to empty dict. "
    "Custom search algorithms may ignore this. "
    "Expects a path to a python script containing a `get_config` function. ",
)
def experiment(run_or_experiment, config):
    """Launch a Tune experiment from a config file."""
    from raylab.utils.dynamic_import import import_module_from_path

    if config:
        config = import_module_from_path(config).get_config()

    return run_or_experiment, config, {}



"""NN architecture used in Deep Deterministic Policy Gradients."""
from dataclasses import dataclass, field

from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box
from nnrl.nn.actor.deterministic import DeterministicActor
from nnrl.nn.critic import ActionValueCritic
from torch import nn


@dataclass
class ActorSpec(DeterministicActor.spec_cls):
    """Specifications for policy, behavior, and target policy."""

    separate_target_policy: bool = field(default=False, init=False, repr=False)
    smooth_target_policy: bool = field(default=True, init=False, repr=False)


CriticSpec = ActionValueCritic.spec_cls


def default_actor() -> ActorSpec:
    spec = ActorSpec()
    spec.network.units = (256, 256)
    spec.network.activation = "ReLU"
    spec.network.layer_norm = False
    spec.network.norm_beta = 1.2
    spec.separate_behavior = False
    spec.target_gaussian_sigma = 0.3
    spec.initializer = {}
    return spec


def default_critic() -> CriticSpec:
    spec = CriticSpec()
    spec.encoder.units = (256, 256)
    spec.encoder.activation = "ReLU"
    spec.encoder.delay_action = False
    spec.double_q = True
    spec.parallelize = True
    spec.initializer = {}
    return spec


@dataclass
class SOPSpec(DataClassJsonMixin):
    """Specifications for SOP modules.

    Args:
        actor: Specifications for policy, behavior, and target policy
        critic: Specifications for action-value estimators
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments. Overrides actor and critic initializer
            specifications.
    """

    actor: ActorSpec = field(default_factory=default_actor)
    critic: CriticSpec = field(default_factory=default_critic)
    initializer: dict = field(default_factory=dict)

    def __post_init__(self):
        # Top-level initializer options take precedence over individual
        # component's options
        if self.initializer:
            self.actor.initializer = self.initializer
            self.critic.initializer = self.initializer


class SOP(nn.Module):
    """NN module for SOP-like algorithms.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for DDPG modules

    Attributes:
        actor (DeterministicPolicy): The deterministic policy to be learned
        behavior (DeterministicPolicy): The policy for exploration
        target_actor (DeterministicPolicy): The policy used for estimating the
            arg max in Q-Learning
        critics (QValueEnsemble): The action-value estimators to be learned
        target_critics (QValueEnsemble): The action-value estimators used for
            bootstrapping in Q-Learning
        spec_cls: Expected class of `spec` init argument
    """

    # pylint:disable=abstract-method
    spec_cls = SOPSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: SOPSpec):
        super().__init__()
        # Build actor
        actor = DeterministicActor(obs_space, action_space, spec.actor)
        self.actor = actor.policy
        self.behavior = actor.behavior
        self.target_actor = actor.target_policy

        main, target = set(self.actor.parameters()), set(self.target_actor.parameters())
        symdiff = main.symmetric_difference(target)
        assert not symdiff, f"Main and target policy must be the same: {symdiff}"

        # Build critic
        critic = ActionValueCritic(obs_space, action_space, spec.critic)
        self.critics = critic.q_values
        self.target_critics = critic.target_q_values



"""Base for all PyTorch policies."""
import textwrap
from typing import Dict, List, Optional, Set, Tuple, Type, Union

import torch
from gym.spaces import Space
from nnrl.utils import convert_to_tensor
from ray.rllib import Policy, SampleBatch
from ray.rllib.evaluation.episode import MultiAgentEpisode
from ray.rllib.models.modelv2 import flatten, restore_original_dimensions
from ray.rllib.policy.view_requirement import ViewRequirement
from ray.rllib.utils import override
from ray.rllib.utils.torch_ops import convert_to_non_torch_type, convert_to_torch_tensor
from ray.rllib.utils.typing import ModelGradients, TensorType
from ray.tune.logger import pretty_print
from torch import Tensor, nn

from raylab.options import RaylabOptions, configure, option

from .action_dist import BaseActionDist
from .compat import WrapRawModule
from .modules import get_module
from .optimizer_collection import OptimizerCollection


@configure
@option("env", default=None)
@option("env_config/", allow_unknown_subkeys=True)
@option("explore", default=True)
@option(
    "exploration_config/", allow_unknown_subkeys=True, override_all_if_type_changes=True
)
@option("framework", default="torch")
@option("gamma", default=0.99)
@option("num_workers", default=0)
@option("seed", default=None)
@option("worker_index", default=0)
@option("normalize_actions", default=True)
@option("clip_actions", default=False)
@option(
    "module/",
    help="Type and config of the PyTorch NN module.",
    allow_unknown_subkeys=True,
    override_all_if_type_changes=True,
)
@option(
    "optimizer/",
    help="Config dict for PyTorch optimizers.",
    allow_unknown_subkeys=True,
)
@option("compile", False, help="Whether to optimize the policy's backend")
class TorchPolicy(Policy):
    """A Policy that uses PyTorch as a backend.

    Attributes:
        observation_space: Space of possible observation inputs
        action_space: Space of possible action outputs
        config: Policy configuration
        dist_class: Action distribution class for computing actions. Must be set
            by subclasses before calling `__init__`.
        device: Device in which the parameter tensors reside. All input samples
            will be converted to tensors and moved to this device
        module: The policy's neural network module. Should be compilable to
            TorchScript
        optimizers: The optimizers bound to the neural network (or submodules)
        options: Configuration object for this class
    """

    observation_space: Space
    action_space: Space
    dist_class: Type[BaseActionDist]
    config: dict
    global_config: dict
    device: torch.device
    model: WrapRawModule
    module: nn.Module
    optimizers: OptimizerCollection
    options: RaylabOptions = RaylabOptions()

    def __init__(self, observation_space: Space, action_space: Space, config: dict):
        # Allow subclasses to set `dist_class` before calling init
        action_dist: Optional[Type[BaseActionDist]] = getattr(self, "dist_class", None)
        super().__init__(observation_space, action_space, self._build_config(config))

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.module = self._make_module(observation_space, action_space, self.config)
        self.module.to(self.device)
        self.model = WrapRawModule(
            self.observation_space,
            self.action_space,
            self.module,
            num_outputs=1,
            model_config=self.config["module"],
        )

        self.optimizers = self._make_optimizers()

        # === Policy attributes ===
        self.dist_class: Type[BaseActionDist] = action_dist
        self.dist_class.check_model_compat(self.module)
        self.framework = "torch"  # Needed to create exploration
        self.exploration = self._create_exploration()

    # ==========================================================================
    # PublicAPI
    # ==========================================================================

    @property
    def pull_from_global(self) -> Set[str]:
        """Keys to pull from global configuration.

        Configurations passed down from caller (usually by the trainer) that are
        not under the `policy` config.
        """
        return {
            "env",
            "env_config",
            "explore",
            # "exploration_config",
            "gamma",
            "num_workers",
            "seed",
            "worker_index",
            "normalize_actions",
            "clip_actions",
        }

    def compile(self):
        """Optimize modules with TorchScript.

        Warnings:
            This action cannot be undone.
        """
        self.module = torch.jit.script(self.module)

    @torch.no_grad()
    @override(Policy)
    def compute_actions(
        self,
        obs_batch: Union[List[TensorType], TensorType],
        state_batches: Optional[List[TensorType]] = None,
        prev_action_batch: Union[List[TensorType], TensorType] = None,
        prev_reward_batch: Union[List[TensorType], TensorType] = None,
        info_batch: Optional[Dict[str, list]] = None,
        episodes: Optional[List[MultiAgentEpisode]] = None,
        explore: Optional[bool] = None,
        timestep: Optional[int] = None,
        **kwargs,
    ) -> Tuple[TensorType, List[TensorType], Dict[str, TensorType]]:
        # pylint:disable=too-many-arguments,too-many-locals
        explore = explore if explore is not None else self.config["explore"]
        timestep = timestep if timestep is not None else self.global_timestep

        input_dict = self.lazy_tensor_dict(
            SampleBatch({SampleBatch.CUR_OBS: obs_batch, "is_training": False})
        )
        if prev_action_batch is not None:
            input_dict[SampleBatch.PREV_ACTIONS] = prev_action_batch
        if prev_reward_batch is not None:
            input_dict[SampleBatch.PREV_REWARDS] = prev_reward_batch
        state_batches = convert_to_torch_tensor(state_batches or [], device=self.device)

        # Call the exploration before_compute_actions hook.
        self.exploration.before_compute_actions(timestep=timestep)

        unpacked = unpack_observations(
            input_dict, self.observation_space, self.framework
        )
        state_out = state_batches

        # pylint:disable=not-callable
        action_dist = self.dist_class({"obs": unpacked["obs"]}, self.model)
        # pylint:enable=not-callable
        actions, logp = self.exploration.get_exploration_action(
            action_distribution=action_dist, timestep=timestep, explore=explore
        )
        input_dict[SampleBatch.ACTIONS] = actions

        # Add default and custom fetches.
        extra_fetches = {}
        if logp is not None:
            extra_fetches[SampleBatch.ACTION_PROB] = logp.exp()
            extra_fetches[SampleBatch.ACTION_LOGP] = logp

        return convert_to_non_torch_type((actions, state_out, extra_fetches))

    def _get_default_view_requirements(self):
        # Add extra fetch keys to view requirements so that they're available
        # for training
        return {
            SampleBatch.ACTION_PROB: ViewRequirement(used_for_training=True),
            SampleBatch.ACTION_LOGP: ViewRequirement(used_for_training=True),
            **super()._get_default_view_requirements(),
        }

    @torch.no_grad()
    @override(Policy)
    def compute_log_likelihoods(
        self,
        actions: Union[List[TensorType], TensorType],
        obs_batch: Union[List[TensorType], TensorType],
        state_batches: Optional[List[TensorType]] = None,
        prev_action_batch: Optional[Union[List[TensorType], TensorType]] = None,
        prev_reward_batch: Optional[Union[List[TensorType], TensorType]] = None,
        actions_normalized: bool = True,
    ) -> TensorType:
        # pylint:disable=too-many-arguments
        input_dict = self.lazy_tensor_dict(
            SampleBatch({SampleBatch.CUR_OBS: obs_batch, SampleBatch.ACTIONS: actions})
        )
        if prev_action_batch is not None:
            input_dict[SampleBatch.PREV_ACTIONS] = prev_action_batch
        if prev_reward_batch is not None:
            input_dict[SampleBatch.PREV_REWARDS] = prev_reward_batch

        dist_inputs, _ = self.module(
            unpack_observations(input_dict, self.observation_space, self.framework),
            state_batches,
            self.convert_to_tensor([1]),
        )
        # pylint:disable=not-callable
        action_dist = self.dist_class(dist_inputs, self.module)
        # pylint:enable=not-callable
        log_likelihoods = action_dist.logp(input_dict[SampleBatch.ACTIONS])
        return log_likelihoods

    @override(Policy)
    def postprocess_trajectory(
        self, sample_batch, other_agent_batches=None, episode=None
    ):
        if not self.config["env_config"].get("time_aware", False):
            hit_limit = sample_batch[SampleBatch.INFOS][-1].get("TimeLimit.truncated")
            env_done = sample_batch[SampleBatch.DONES][-1]
            sample_batch[SampleBatch.DONES][-1] = False if hit_limit else env_done
        return sample_batch

    @override(Policy)
    def get_weights(self) -> dict:
        return {
            "module": convert_to_non_torch_type(self.module.state_dict()),
            # Optimizer state dicts don't store tensors, only ids
            "optimizers": self.optimizers.state_dict(),
        }

    @override(Policy)
    def set_weights(self, weights: dict):
        self.module.load_state_dict(
            convert_to_torch_tensor(weights["module"], device=self.device)
        )
        # Optimizer state dicts don't store tensors, only ids
        self.optimizers.load_state_dict(weights["optimizers"])

    def convert_to_tensor(self, arr) -> Tensor:
        """Convert an array to a PyTorch tensor in this policy's device.

        Args:
            arr (array_like): object which can be converted using `np.asarray`
        """
        return convert_to_tensor(arr, self.device)

    def lazy_tensor_dict(self, sample_batch: SampleBatch) -> SampleBatch:
        """Convert a sample batch into a dictionary of lazy tensors.

        The sample batch is wrapped with a UsageTrackingDict to convert array-
        likes into tensors upon querying.

        Args:
            sample_batch: the sample batch to convert

        Returns:
            A dictionary which intercepts key queries to lazily convert arrays
            to tensors.
        """
        tensor_batch = sample_batch.copy(shallow=True)
        tensor_batch.set_get_interceptor(self.convert_to_tensor)
        return tensor_batch

    def __repr__(self):
        name = self.__class__.__name__
        args = [f"{self.observation_space},", f"{self.action_space},"]

        config = pretty_print(self.config).rstrip("\n")
        if "\n" in config:
            config = textwrap.indent(config, " " * 2)
            config = "{\n" + config + "\n}"

            args += [config]
            args_repr = "\n".join(args)
            args_repr = textwrap.indent(args_repr, " " * 2)
            constructor = f"{name}(\n{args_repr}\n)"
        else:
            args += [config]
            args_repr = " ".join(args[1:-1])
            constructor = f"{name}({args_repr})"
        return constructor

    def apply_gradients(self, gradients: ModelGradients) -> None:
        pass

    def compute_gradients(
        self, postprocessed_batch: SampleBatch
    ) -> Tuple[ModelGradients, Dict[str, TensorType]]:
        pass

    # ==========================================================================
    # InternalAPI
    # ==========================================================================

    def _build_config(self, config: dict) -> dict:
        if not self.options.all_options_set:
            raise RuntimeError(
                f"{type(self).__name__} still has configs to be set."
                " Did you call `configure` as the last decorator?"
            )

        passed = config.get("policy", {}).copy()
        passed.update({k: config[k] for k in self.pull_from_global if k in config})
        new = self.options.merge_defaults_with(passed)
        return new

    @staticmethod
    def _make_module(obs_space: Space, action_space: Space, config: dict) -> nn.Module:
        """Build the PyTorch nn.Module to be used by this policy.

        Args:
            obs_space: the observation space for this policy
            action_space: the action_space for this policy
            config: the user config containing the 'module' key

        Returns:
            A neural network module.
        """
        return get_module(obs_space, action_space, config["module"])

    def _make_optimizers(self) -> OptimizerCollection:
        """Build PyTorch optimizers to use.

        The result will be set as the policy's optimizer collection.

        The user should update the optimizer collection (mutable mapping)
        returned by the base implementation.

        Returns:
            A mapping from names to optimizer instances
        """
        # pylint:disable=no-self-use
        return OptimizerCollection()

    # ==========================================================================
    # Unimplemented Policy methods
    # ==========================================================================

    def get_num_samples_loaded_into_buffer(self, buffer_index: int = 0) -> int:
        pass

    def learn_on_loaded_batch(self, offset: int = 0, buffer_index: int = 0):
        pass

    def load_batch_into_buffer(self, batch: SampleBatch, buffer_index: int = 0) -> int:
        pass

    def export_model(self, export_dir: str, onnx: Optional[int] = None) -> None:
        pass

    def export_checkpoint(self, export_dir):
        pass

    def import_model_from_h5(self, import_file):
        pass


def unpack_observations(input_dict, observation_space: Space, framework: str):
    """Cast observations to original space and add a separate flattened view."""
    restored = input_dict.copy()
    restored["obs"] = restore_original_dimensions(
        input_dict["obs"], observation_space, framework
    )
    if len(input_dict["obs"].shape) > 2:
        restored["obs_flat"] = flatten(input_dict["obs"], framework)
    else:
        restored["obs_flat"] = input_dict["obs"]
    return restored



"""NN architecture used in Normalized Advantage Function."""
from dataclasses import dataclass, field

from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box
from nnrl.nn.actor import DeterministicPolicy, MLPDeterministicPolicy
from nnrl.nn.critic import (
    ForkedQValueEnsemble,
    ForkedVValueEnsemble,
    QValueEnsemble,
    VValueEnsemble,
)
from nnrl.nn.critic.naf_value import NAFQValue
from torch import nn

MLPPolicySpec = MLPDeterministicPolicy.spec_cls


@dataclass
class NAFSpec(DataClassJsonMixin):
    """Specifications for Normalized Advantage Function."""

    policy: MLPPolicySpec = field(default_factory=MLPPolicySpec)
    separate_behavior: bool = False
    double_q: bool = True
    parallelized: bool = False


class NAF(nn.Module):
    """Neural network architecture for Normalized Advantage Function."""

    # pylint:disable=abstract-method
    spec_cls = NAFSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: NAFSpec):
        super().__init__()

        def make_policy():
            return MLPDeterministicPolicy(obs_space, action_space, spec.policy)

        policies = [make_policy() for _ in range(1 + spec.double_q)]
        self.actor: DeterministicPolicy = policies[0]
        self.behavior: DeterministicPolicy = self.actor
        if spec.separate_behavior:
            self.behavior = make_policy()
            self.behavior.load_state_dict(self.actor.state_dict())

        ensemble_cls = ForkedQValueEnsemble if spec.parallelized else QValueEnsemble
        self.critics: QValueEnsemble = ensemble_cls(
            [NAFQValue(action_space, pol) for pol in policies]
        )

        ensemble_cls = ForkedVValueEnsemble if spec.parallelized else VValueEnsemble
        self.vcritics: VValueEnsemble = ensemble_cls([q.v_value for q in self.critics])
        self.target_vcritics: VValueEnsemble = ensemble_cls(
            [NAFQValue(action_space, make_policy()).v_value for _ in self.critics]
        )
        self.target_vcritics.load_state_dict(self.vcritics.state_dict())



"""Abstract base classes for losses."""
from abc import ABCMeta, abstractmethod
from typing import Dict, Tuple

from nnrl.types import TensorDict
from torch import Tensor

from raylab.utils.dictionaries import get_keys
from raylab.utils.types import StatDict

__all__ = [
    "Dict",
    "Tuple",
    "Tensor",
    "Loss",
]


class Loss(metaclass=ABCMeta):
    """Base interface for all loss classes.

    Attributes:
        batch_keys: the fiels in the tensor batch which will be accessed when
            called. Needed for converting the appropriate inputs to tensors
            externally.
    """

    batch_keys: Tuple[str, ...]

    @abstractmethod
    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        """Computes the loss function and stats dict for the given batch.

        Subclasses should override this to implement their respective loss
        functions.

        Args:
            batch: a dictionary of input tensors

        Returns:
            A tuple with the loss tensor (not necessarily scalar) and statistics
        """

    def compile(self):
        """Optimize this loss function's implementation if supported.

        Subclasses should use TorchScript to build a static computation graph,
        although this may not be possible in some cases.
        """

    def unpack_batch(self, batch: TensorDict) -> Tuple[Tensor, ...]:
        """Returns the batch tensors corresponding to the batch keys.

        Tensors are returned in the same order `batch_keys` is defined.

        Args:
            batch: Dictionary of input tensors

        Returns:
            A tuple of tensors corresponding to each key in `batch_keys`
        """
        return tuple(get_keys(batch, *self.batch_keys))



"""Architecture used in Trust Region Policy Optimization."""
from dataclasses import dataclass, field

from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box
from nnrl.nn.actor import StochasticActor
from nnrl.nn.critic import MLPVValue
from torch import nn

ActorSpec = StochasticActor.spec_cls
CriticSpec = MLPVValue.spec_cls


@dataclass
class TRPOSpec(DataClassJsonMixin):
    # pylint:disable=missing-class-docstring
    actor: ActorSpec = field(default_factory=ActorSpec)
    critic: CriticSpec = field(default_factory=CriticSpec)


class TRPO(nn.Module):
    # pylint:disable=missing-class-docstring
    spec_cls = TRPOSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: TRPOSpec):
        super().__init__()
        self._make_actor(obs_space, action_space, spec.actor)
        self._make_critic(obs_space, spec.critic)

    def _make_actor(self, obs_space: Box, action_space: Box, spec: ActorSpec):
        actor = StochasticActor(obs_space, action_space, spec)
        self.actor = actor.policy

    def _make_critic(self, obs_space: Box, spec: CriticSpec):
        self.critic = MLPVValue(obs_space, spec)



# pylint:disable=missing-docstring
from ray.rllib.utils import override
from ray.tune.logger import VALID_SUMMARY_TYPES, Logger
from ray.tune.result import TIME_TOTAL_S, TIMESTEPS_TOTAL, TRAINING_ITERATION
from ray.tune.utils.util import flatten_dict
from torch.utils.tensorboard import SummaryWriter


class TorchTBLogger(Logger):
    """Tensorboard logger with no tensorflow dependency."""

    @override(Logger)
    def _init(self):
        self._file_writer = SummaryWriter(self.logdir)

    @override(Logger)
    def on_result(self, result):
        tmp = result.copy()
        elim = ["config", "pid", "timestamp", "done", TIME_TOTAL_S, TRAINING_ITERATION]
        for k in elim:
            if k in tmp:
                del tmp[k]

        step = result.get(TIMESTEPS_TOTAL) or result[TRAINING_ITERATION]
        tmp.update({TRAINING_ITERATION: result[TRAINING_ITERATION]})
        for key, val in flatten_dict(tmp).items():
            if isinstance(val, tuple(VALID_SUMMARY_TYPES)):
                self._file_writer.add_scalar("/".join(["ray", "tune", key]), val, step)

        self._file_writer.flush()

    @override(Logger)
    def flush(self):
        self._file_writer.flush()

    @override(Logger)
    def close(self):
        self._file_writer.close()



"""Utilities for handling experiment results."""
import os.path as osp
import pickle
import warnings

from ray.rllib.utils import merge_dicts

from raylab.agents.registry import get_agent_cls


def get_agent_from_checkpoint(checkpoint, agent_name, env=None, **config_kwargs):
    """Instatiate and restore agent class from checkpoint."""
    config = get_config_from_checkpoint(checkpoint, **config_kwargs)
    agent_cls = get_agent_cls(agent_name)

    agent = agent_cls(env=env, config=config)
    agent.restore(checkpoint)
    return agent


def get_config_from_checkpoint(checkpoint, use_eval_config=True, config_overrides=None):
    """Find and load configuration for checkpoint file."""
    config = {}
    # Load configuration from file
    config_dir = osp.dirname(checkpoint)
    config_path = osp.join(config_dir, "params.pkl")
    if not osp.exists(config_path):
        config_path = osp.join(config_dir, "../params.pkl")
    if not osp.exists(config_path):
        raise ValueError(
            "Could not find params.pkl in either the checkpoint dir or "
            "its parent directory."
        )
    with open(config_path, "rb") as file:
        config = pickle.load(file)

    if use_eval_config:
        if "evaluation_config" not in config:
            warnings.warn("Evaluation agent requested but none in config.")
        eval_conf = config.get("evaluation_config", {})
        config = merge_dicts(config, eval_conf)

    if config_overrides:
        config = merge_dicts(config, config_overrides)
    return config



"""Trainer and configuration for MBPO."""
from raylab.agents import Trainer
from raylab.agents.model_based import ModelBasedMixin
from raylab.agents.off_policy import OffPolicyMixin
from raylab.options import configure

from .policy import MBPOTorchPolicy


@configure
@OffPolicyMixin.add_options
class MBPOTrainer(ModelBasedMixin, OffPolicyMixin, Trainer):
    """Model-based trainer using SAC for policy improvement."""

    _name = "MBPO"
    _policy_class = MBPOTorchPolicy



# pylint:disable=missing-module-docstring
import gym
import numpy as np

from .utils import assert_flat_box_space, ignore_rightmost_variables


class AddRelativeTimestep(gym.ObservationWrapper):
    """
    Adds the relative timestep (normalized to the range [0, 1]) to the
    observations of environments with observation spaces of type gym.spaces.Box.
    """

    def __init__(self, env=None):
        assert_flat_box_space(env.observation_space, self)
        super().__init__(env)

        _env = env
        while hasattr(_env, "env"):
            if isinstance(_env, gym.wrappers.TimeLimit):
                break
            _env = _env.env
        self._env = _env

        self.observation_space = gym.spaces.Box(
            low=np.append(self.observation_space.low, 0.0),
            high=np.append(self.observation_space.high, 1.0),
            dtype=self.observation_space.dtype,
        )

        if hasattr(self.env, "reward_fn"):
            self.reward_fn = ignore_rightmost_variables(self.env.reward_fn, 1)

        if hasattr(self.env, "termination_fn"):

            def termination_fn(state, action, next_state):
                env_done = self.env.termination_fn(
                    state[..., :-1], action, next_state[..., :-1]
                )
                timeout = next_state[..., -1] >= 1.0
                return timeout | env_done

            self.termination_fn = termination_fn

    def observation(self, observation):
        # pylint:disable=protected-access
        relative_time = self._env._elapsed_steps / self._env._max_episode_steps
        return np.append(observation, relative_time)



"""Losses for model aware action gradient estimation."""
from typing import Tuple, Union

import torch
from nnrl.nn.actor import DeterministicPolicy
from nnrl.nn.critic import QValueEnsemble, VValue
from nnrl.nn.model import SME, StochasticModel
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from torch import Tensor

from raylab.utils.types import StatDict

from .abstract import Loss
from .mixins import EnvFunctionsMixin, UniformModelPriorMixin
from .utils import dist_params_stats


class MAGE(EnvFunctionsMixin, UniformModelPriorMixin, Loss):
    """Loss function for Model-based Action-Gradient-Estimator.

    Args:
        critics: Q-value estimators
        policy: deterministic policy for current state
        target_critic: V-value estimator for the next state
        models: ensemble of stochastic models

    Attributes:
        gamma: discount factor
        lambd: weighting factor for TD-error regularization
    """

    batch_keys = (SampleBatch.CUR_OBS,)
    gamma: float = 0.99
    lambd: float = 0.05

    def __init__(
        self,
        critics: QValueEnsemble,
        policy: DeterministicPolicy,
        target_critic: VValue,
        models: Union[StochasticModel, SME],
    ):
        super().__init__()
        self.critics = critics
        self.policy = policy
        self.target_critic = target_critic
        if isinstance(models, StochasticModel):
            models = SME([models])
        self.models = models

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        self.check_env_fns()

        obs = batch[SampleBatch.CUR_OBS]
        action = self.policy(obs)
        next_obs, dist_params = self.transition(obs, action)

        delta = self.temporal_diff_error(obs, action, next_obs)
        grad_loss = self.gradient_loss(delta, action)
        td_reg = self.temporal_diff_loss(delta)
        loss = grad_loss + self.lambd * td_reg

        info = {
            "loss(critics)": loss.item(),
            "loss(MAGE)": grad_loss.item(),
            "loss(TD)": td_reg.item(),
        }
        info.update(dist_params_stats(dist_params, name="model"))
        return loss, info

    def transition(self, obs: Tensor, action: Tensor) -> Tuple[Tensor, TensorDict]:
        # pylint:disable=missing-function-docstring
        model, _ = self.sample_model()
        dist_params = model(obs, action)
        next_obs, _ = model.rsample(dist_params)
        return next_obs, dist_params

    def temporal_diff_error(
        self,
        obs: Tensor,
        action: Tensor,
        next_obs: Tensor,
    ) -> Tensor:
        """Returns the temporal difference error."""
        reward = self._env.reward(obs, action, next_obs)  # (*,)
        done = self._env.termination(obs, action, next_obs)  # (*,)
        next_val = self.target_critic(next_obs)  # (*,)
        target = torch.where(done, reward, reward + self.gamma * next_val)  # (*,)

        values = self.critics(obs, action)  # [(*,)] * N
        return torch.stack([target - v for v in values], dim=-1)  # (*, N)

    @staticmethod
    def gradient_loss(delta: Tensor, action: Tensor) -> Tensor:
        """Returns the action gradient loss for the Q-value function."""
        (action_gradient,) = torch.autograd.grad(delta.sum(), action, create_graph=True)
        return torch.sum(action_gradient ** 2, dim=-1).mean()

    @staticmethod
    def temporal_diff_loss(delta: Tensor) -> Tensor:
        """Returns the temporal difference loss for the Q-value function."""
        return torch.sum(delta ** 2, dim=-1).mean()



# pylint:disable=missing-module-docstring
import textwrap
from typing import Any

from gym.spaces import Box, Space


def assert_flat_box_space(space: Space, obj: Any):
    """Check that `space` is a 1D Box space.

    Args:
        env: Gym space instance
        obj: Object associated with the space

    Raises:
        AssertionError: If `space` is not a 1D Box space
    """
    assert isinstance(space, Box), (
        f"{type(obj).__name__} only compatible with 1D Box spaces."
        f" Got: {type(space).__name__}."
    )
    assert len(space.shape) == 1, (
        f"{type(obj).__name__} only compatible with 1D Box spaces."
        f" Got Box space with shape {space.shape}"
    )


def check_redundant_size_compat(size: int, space: Box):
    """Check that `size` is less or equal to the size of the observations."""
    assert size <= space.shape[0], textwrap.dedent(
        f"""\
        Can only compute redundant variables from a subset of observation features.
        `size` must be less or equal to the size of the base env observations.
        Got '{size}', whereas the observation shape is {space.shape}
        """
    )


def ignore_rightmost_variables(func: callable, size: int) -> callable:
    """Wrap base env reward/termination function to ignore added variables.

    Args:
        func: Callable for reward/termination function
        size: Number of irrelevant/redundant variables
    """

    def env_fn(state, action, next_state):
        return func(state[..., :-size], action, next_state[..., :-size])

    return env_fn



"""Trainer and configuration for TRPO."""
from raylab.agents.trainer import Trainer

from .policy import TRPOTorchPolicy


class TRPOTrainer(Trainer):
    """Single agent trainer for TRPO."""

    _name = "TRPO"
    _policy_class = TRPOTorchPolicy



"""Custom Replay Buffers based on RLlibs's implementation."""
from dataclasses import dataclass
from typing import Dict, Optional, Tuple, Union

import numpy as np
from gym.spaces import Space
from ray.rllib import SampleBatch


@dataclass
class ReplayField:
    """Storage specification for ReplayBuffer data."""

    name: str
    shape: tuple = ()
    dtype: np.dtype = np.float32


class NumpyReplayBuffer:
    """Replay buffer as a dict of ndarrays.

    Returns a SampleBatch object when queried for samples.

    Args:
        obs_space: observation space
        action_space: action space
        size: max number of transitions to store in the buffer.
            When the bufferoverflows the old memories are dropped.

    Attributes:
        fields (:obj:`tuple` of :obj:`ReplayField`): storage fields
            specification
        compute_stats: Whether to track mean and stddev for normalizing
            observations
    """

    # pylint:disable=too-many-instance-attributes
    compute_stats: bool = False

    def __init__(self, obs_space: Space, action_space: Space, size: int):
        self._maxsize = size
        self.fields = (
            ReplayField(
                SampleBatch.CUR_OBS, shape=obs_space.shape, dtype=obs_space.dtype
            ),
            ReplayField(
                SampleBatch.ACTIONS, shape=action_space.shape, dtype=action_space.dtype
            ),
            ReplayField(SampleBatch.REWARDS, shape=(), dtype=np.float32),
            ReplayField(
                SampleBatch.NEXT_OBS, shape=obs_space.shape, dtype=obs_space.dtype
            ),
            ReplayField(SampleBatch.DONES, shape=(), dtype=np.bool),
        )
        self._storage = {}
        self._build_buffers(*self.fields)
        self._next_idx = 0
        self._curr_size = 0
        self._rng = np.random.default_rng()
        self._obs_stats: Optional[Tuple[np.ndarray, np.ndarray]] = None

    def __len__(self) -> int:
        return self._curr_size

    def add_fields(self, *fields: ReplayField):
        """Add fields to the replay buffer and build the corresponding storage."""
        new_names = {f.name for f in fields}
        assert len(new_names) == len(fields), "Field names must be unique"

        conflicts = new_names.intersection({f.name for f in self.fields})
        assert not conflicts, f"{conflicts} are already in buffer"

        self.fields = self.fields + fields
        self._build_buffers(*fields)

    def _build_buffers(self, *fields: ReplayField):
        storage = self._storage
        size = self._maxsize
        for field in fields:
            storage[field.name] = np.empty((size,) + field.shape, dtype=field.dtype)

    def __getitem__(
        self, index: Union[int, np.ndarray, slice]
    ) -> Dict[str, np.ndarray]:
        batch = {f.name: self._storage[f.name][index] for f in self.fields}
        for key in SampleBatch.CUR_OBS, SampleBatch.NEXT_OBS:
            batch[key] = self.normalize(batch[key])
        return batch

    def normalize(self, obs: np.ndarray) -> np.ndarray:
        """Normalize observation using the stored mean and stddev."""
        obs = np.asarray(obs)
        if not self.compute_stats:
            return obs

        if not self._obs_stats:
            self.update_obs_stats()

        mean, std = self._obs_stats
        return (obs - mean) / std

    def update_obs_stats(self):
        """Compute mean and standard deviation for observation normalization.

        Subsequent batches sampled from this buffer will use these statistics to
        normalize the current and next observation fields.
        """
        if len(self) == 0:
            self._obs_stats = (0, 1)
        else:
            cur_obs = self._storage[SampleBatch.CUR_OBS][: len(self)]
            mean = np.mean(cur_obs, axis=0)
            std = np.std(cur_obs, axis=0)
            std[std < 1e-12] = 1.0
            self._obs_stats = (mean, std)

    def seed(self, seed: int = None):
        """Seed the random number generator for sampling minibatches."""
        self._rng = np.random.default_rng(seed)

    def add(self, samples: SampleBatch):
        """Add a SampleBatch to storage.

        Optimized to avoid several queries for large sample batches.

        Args:
            samples: The sample batch
        """
        if samples.count >= self._maxsize:
            samples = samples.slice(samples.count - self._maxsize, None)
            end_idx = 0
            assign = [(slice(0, self._maxsize), samples)]
        else:
            start_idx = self._next_idx
            end_idx = (self._next_idx + samples.count) % self._maxsize
            if end_idx < start_idx:
                tailcount = self._maxsize - start_idx
                assign = [
                    (slice(start_idx, None), samples.slice(0, tailcount)),
                    (slice(end_idx), samples.slice(tailcount, None)),
                ]
            else:
                assign = [(slice(start_idx, end_idx), samples)]

        for field in self.fields:
            for slc, smp in assign:
                self._storage[field.name][slc] = smp[field.name]

        self._next_idx = end_idx
        self._curr_size = min(self._curr_size + samples.count, self._maxsize)
        self._obs_stats = None

    def sample(self, batch_size: int) -> SampleBatch:
        """Transition batch uniformly sampled with replacement."""
        return SampleBatch(self[self.sample_idxes(batch_size)])

    def sample_idxes(self, batch_size: int) -> np.ndarray:
        """Get random transition indexes uniformly sampled with replacement."""
        return self._rng.integers(self._curr_size, size=batch_size)

    def all_samples(self) -> SampleBatch:
        """All stored transitions."""
        return SampleBatch(self[: len(self)])

    def state_dict(self) -> dict:
        return {"obs_stats": self._obs_stats}

    def load_state_dict(self, state: dict):
        self._obs_stats = state["obs_stats"]



# pylint:disable=missing-docstring
# pylint: enable=missing-docstring
from abc import abstractmethod

from raylab.utils.adaptive_kl import AdaptiveKLCoeffSpec


class AdaptiveKLCoeffMixin:
    """Adds adaptive KL penalty as in PPO."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._kl_coeff_spec = AdaptiveKLCoeffSpec(**self.config["kl_schedule"])

    def update_kl_coeff(self, sample_batch):
        """Update KL penalty based on observed divergence between successive policies.

        Arguments:
            sample_batch (SampleBatch): batch of data gathered by the old policy

        Returns:
            A dictionary with a single entry corresponding to the average KL divergence.
        """
        kl_div = self._kl_divergence(sample_batch)
        self._kl_coeff_spec.adapt(kl_div)
        return {"policy_kl_div": kl_div}

    @property
    def curr_kl_coeff(self):
        """Return current KL coefficient."""
        return self._kl_coeff_spec.curr_coeff

    @abstractmethod
    def _kl_divergence(self, sample_batch):
        """Compute the empirical average KL divergence between new and old policies.

        Arguments:
            sample_batch (SampleBatch): batch of data gathered by the old policy

        Returns:
            A scalar equal to the average KL divergence between new and old policies.
        """



# pylint:disable=missing-module-docstring
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, Union

from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from ray.rllib.evaluation.episode import MultiAgentEpisode
from ray.rllib.utils.typing import TensorType

from raylab.options import option
from raylab.utils.replay_buffer import NumpyReplayBuffer

from .stats import learner_stats


def off_policy_options(cls: type) -> type:
    """Decorator to add default off-policy options used by OffPolicyMixin."""
    buffer_size = option(
        "buffer_size",
        default=int(1e4),
        help="""Size (number of transitions) of the replay buffer.""",
    )
    std_obs = option(
        "std_obs",
        default=False,
        help="Wheter to normalize replayed observations by the empirical mean and std.",
    )
    improvement_steps = option(
        "improvement_steps",
        default=1,
        help="""Policy improvement steps on each call to `learn_on_batch`.

        Example:
            With a 'rollout_fragment_length' of 1 and 'policy/improvement_steps' equal
            to 10,will perform 10 policy updates with minibatch size 'policy/batch_size'
            per environment step.
        """,
    )
    batch_size = option(
        "batch_size",
        default=128,
        help="Size of replay buffer batches sampled on each call to `improve_policy`.",
    )

    options = [buffer_size, std_obs, improvement_steps, batch_size]
    for opt in options:
        cls = opt(cls)

    return cls


class OffPolicyMixin(ABC):
    """Adds a replay buffer and standard procedures for `learn_on_batch`."""

    replay: NumpyReplayBuffer

    def build_replay_buffer(self):
        """Construct the experience replay buffer.

        Should be called by subclasses on init.
        """
        self.replay = NumpyReplayBuffer(
            self.observation_space, self.action_space, self.config["buffer_size"]
        )
        self.replay.seed(self.config["seed"])
        self.replay.compute_stats = self.config["std_obs"]

    @learner_stats
    def learn_on_batch(self, samples: SampleBatch):
        """Run one logical iteration of training.

        Returns:
            An info dict from this iteration.
        """
        self.add_to_buffer(samples)

        info = {}
        info.update(self.get_exploration_info())

        for _ in range(int(self.config["improvement_steps"])):
            batch = self.replay.sample(self.config["batch_size"])
            batch = self.lazy_tensor_dict(batch)
            info.update(self.improve_policy(batch))

        return info

    def add_to_buffer(self, samples: SampleBatch):
        """Add sample batch to replay buffer"""
        self.replay.add(samples)

    @abstractmethod
    def improve_policy(self, batch: TensorDict) -> dict:
        """Run one step of Policy Improvement."""

    def compute_actions(
        self,
        obs_batch: Union[List[TensorType], TensorType],
        state_batches: Optional[List[TensorType]] = None,
        prev_action_batch: Union[List[TensorType], TensorType] = None,
        prev_reward_batch: Union[List[TensorType], TensorType] = None,
        info_batch: Optional[Dict[str, list]] = None,
        episodes: Optional[List[MultiAgentEpisode]] = None,
        explore: Optional[bool] = None,
        timestep: Optional[int] = None,
        **kwargs
    ) -> Tuple[TensorType, List[TensorType], Dict[str, TensorType]]:
        # pylint:disable=too-many-arguments
        obs_batch = self.replay.normalize(obs_batch)
        return super().compute_actions(
            obs_batch,
            state_batches=state_batches,
            prev_action_batch=prev_action_batch,
            prev_reward_batch=prev_reward_batch,
            info_batch=info_batch,
            episodes=episodes,
            explore=explore,
            timestep=timestep,
            **kwargs,
        )

    def compute_log_likelihoods(
        self,
        actions: Union[List[TensorType], TensorType],
        obs_batch: Union[List[TensorType], TensorType],
        state_batches: Optional[List[TensorType]] = None,
        prev_action_batch: Optional[Union[List[TensorType], TensorType]] = None,
        prev_reward_batch: Optional[Union[List[TensorType], TensorType]] = None,
    ) -> TensorType:
        # pylint:disable=too-many-arguments
        obs_batch = self.replay.normalize(obs_batch)
        return super().compute_log_likelihoods(
            actions=actions,
            obs_batch=obs_batch,
            state_batches=state_batches,
            prev_action_batch=prev_action_batch,
            prev_reward_batch=prev_reward_batch,
        )

    def get_weights(self) -> dict:
        state = super().get_weights()
        state["replay"] = self.replay.state_dict()
        return state

    def set_weights(self, weights: dict):
        self.replay.load_state_dict(weights["replay"])
        super().set_weights({k: v for k, v in weights.items() if k != "replay"})

    @staticmethod
    def add_options(policy_cls: type) -> type:
        """Decorator to add default off-policy options used by OffPolicyMixin."""
        return off_policy_options(policy_cls)



"""Base implementations for all exploration strategies."""
import textwrap
from abc import ABCMeta, abstractmethod
from typing import Optional

from ray.rllib.utils.exploration import Exploration
from torch import nn

Model = Optional[nn.Module]


class IncompatibleExplorationError(Exception):
    """Exception raised for incompatible exploration and NN module.

    Args:
        exp_cls: Exploration class
        module: NN module
        err: AssertionError explaining the reason why exploration and module are
            incompatible

    Attributes:
        message: Human-readable text explaining what caused the incompatibility
    """

    def __init__(self, exp_cls: type, module: Model, err: Exception):
        # pylint:disable=unused-argument
        msg = f"""\
        Exploration type {exp_cls} is incompatible with NN module of type
        {type(module)}. Reason:
            {err}
        """
        super().__init__(textwrap.dedent(msg))
        self.message = msg


class BaseExploration(Exploration, metaclass=ABCMeta):
    """Base class for exploration objects."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.check_model_compat(self.model.module)
        except AssertionError as err:
            raise IncompatibleExplorationError(
                type(self), self.model.module, err
            ) from err

    @classmethod
    @abstractmethod
    def check_model_compat(cls, model: Model):
        """Assert the given NN module is compatible with the exploration.

        Raises:
            IncompatibleDistClsError: If `model` is incompatible with the
                exploration class
        """



"""Collection of custom RLlib Policy classes."""
from .kl_coeff_mixin import AdaptiveKLCoeffMixin
from .model_based import *
from .optimizer_collection import OptimizerCollection
from .stats import learner_stats
from .torch_policy import TorchPolicy



"""Modularized losses/procedures for RL algorithms."""
from .abstract import Loss
from .dyna import DynaQLearning
from .isfv_iteration import ISFittedVIteration, ISSoftVIteration
from .mage import MAGE
from .maximum_entropy import MaximumEntropyDual
from .mle import MaximumLikelihood
from .policy_gradient import (
    ActionDPG,
    DeterministicPolicyGradient,
    ReparameterizedSoftPG,
)
from .q_learning import FittedQLearning
from .svg import OneStepSoftSVG, OneStepSVG, TrajectorySVG

__all__ = [
    "Loss",
    "DynaQLearning",
    "FittedQLearning",
    "ActionDPG",
    "DeterministicPolicyGradient",
    "ISFittedVIteration",
    "ISSoftVIteration",
    "MaximumEntropyDual",
    "MAGE",
    "MaximumLikelihood",
    "OneStepSVG",
    "OneStepSoftSVG",
    "ReparameterizedSoftPG",
    "TrajectorySVG",
]



"""Network and configurations for model-based TD3 algorithms."""
from dataclasses import dataclass, field

from gym.spaces import Box
from nnrl.nn.model import EnsembleSpec, build_ensemble

from .sop import SOP, ActorSpec, CriticSpec, SOPSpec


def default_models() -> EnsembleSpec:
    """Model configurations used in MAGE paper for Mujoco environments."""
    spec = EnsembleSpec()
    spec.network.units = (512,) * 4
    spec.network.activation = "Swish"
    spec.network.delay_action = False
    spec.network.fix_logvar_bounds = True
    spec.network.input_dependent_scale = True
    spec.residual = True
    spec.ensemble_size = 8
    spec.parallelize = True
    return spec


def default_actor() -> ActorSpec:
    """Actor configurations used in MAGE paper for Mujoco environments."""
    spec = ActorSpec()
    spec.network.units = 2 * (284,)
    spec.network.activation = "ReLU"
    spec.network.layer_norm = False
    spec.network.norm_beta = 1.2
    spec.separate_behavior = False
    spec.target_gaussian_sigma = 0.3
    return spec


def default_critic() -> CriticSpec:
    """Critic configurations used in MAGE paper for Mujoco environments."""
    spec = CriticSpec()
    spec.encoder.units = 2 * (384,)
    spec.encoder.activation = "ReLU"
    spec.encoder.delay_action = False
    spec.double_q = True
    spec.parallelize = True
    return spec


@dataclass
class MAGESpec(SOPSpec):
    """Specifications for NNs used in MAGE.

    Args:
        model: Specifications for stochastic dynamics model ensemble
        actor: Specifications for policy, behavior, and target policy
        critic: Specifications for action-value estimators
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments. Overrides model, actor, and critic initializer
            specifications.
    """

    model: EnsembleSpec = field(default_factory=default_models)
    actor: ActorSpec = field(default_factory=default_actor)
    critic: CriticSpec = field(default_factory=default_critic)

    def __post_init__(self):
        super().__post_init__()
        if self.initializer:
            self.model.initializer = self.initializer


class MAGE(SOP):
    """NN module for MAGE algorithm

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for NNs used in MAGE

    Attributes:
        model (SME): Stochastic dynamics model ensemble
        actor (DeterministicPolicy): The deterministic policy to be learned
        behavior (DeterministicPolicy): The policy for exploration
        target_actor (DeterministicPolicy): The policy used for estimating the
            arg max in Q-Learning
        critics (QValueEnsemble): The action-value estimators to be learned
        target_critics (QValueEnsemble): The action-value estimators used for
            bootstrapping in Q-Learning
        spec_cls: Expected class of `spec` init argument
    """

    # pylint:disable=abstract-method
    spec_cls = MAGESpec

    def __init__(self, obs_space: Box, action_space: Box, spec: MAGESpec):
        super().__init__(obs_space, action_space, spec)

        self.models = build_ensemble(obs_space, action_space, spec.model)



# pylint:disable=missing-module-docstring
import functools
import logging
import os
import os.path as osp

import click


def initialize_raylab(func):
    """Wrap cli to register raylab's algorithms and environments."""

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        import raylab

        raylab.register_all()

        return func(*args, **kwargs)

    return wrapped


def tune_experiment(func):
    """Transform function into a Click command to launch experiments.

    The wrapped function should return, in order, a trainable class or name, the
    trainable's configutation, and a dict of overrides for the `tune.run` options.
    """

    @click.command()
    @tune_options
    @click.option(
        "--object-store-memory",
        type=int,
        default=int(2e9),
        show_default=True,
        help="The amount of memory (in bytes) to start the object store with. "
        "By default, this is capped at 20GB but can be set higher.",
    )
    @initialize_raylab
    @click.pass_context
    @functools.wraps(func)
    def wrapped(ctx, *args, object_store_memory, tune_kwargs, **kwargs):
        import ray
        from ray import tune
        from ray.rllib.utils import merge_dicts

        trainable, config, tune_overrides = func(*args, **kwargs)
        tune_kwargs = merge_dicts(tune_kwargs, tune_overrides)
        process_tune_kwargs(ctx, **tune_kwargs)

        ray.init(object_store_memory=object_store_memory)
        tune.run(trainable, config=config, **tune_kwargs)

    return wrapped


def tune_options(func):
    """Wrap cli to add and parse arguments for `tune.run`.

    Arguments for `tune.run` will be accessible as a dictionary through the
    `tune_kwargs` kwarg in the wrapped function.

    `tune-log-level` and `--custom-loggers` are already handled by this wrapper.
    No need to process these in the wrapped function.
    """

    @click.option("--name", default=None, help="Name of experiment", required=True)
    @click.option(
        "--local-dir",
        "-l",
        type=click.Path(
            exists=False, file_okay=False, dir_okay=True, resolve_path=True
        ),
        default="data/",
        show_default=True,
        help="",
    )
    @click.option(
        "--num-samples",
        "-n",
        type=int,
        default=1,
        show_default=True,
        help="Number of times to sample from the hyperparameter space. "
        "Defaults to 1. If `grid_search` is provided as an argument, "
        "the grid will be repeated `num_samples` of times.",
    )
    @click.option(
        "--stop",
        "-s",
        type=(str, int),
        multiple=True,
        help="The stopping criteria. "
        "The keys may be any field in the return result of 'train()', "
        "whichever is reached first. Defaults to empty dict.",
    )
    @click.option(
        "--checkpoint-freq",
        type=int,
        default=0,
        show_default=True,
        help="How many training iterations between checkpoints. "
        "A value of 0 disables checkpointing.",
    )
    @click.option(
        "--checkpoint-at-end",
        type=bool,
        default=True,
        show_default=True,
        help="Whether to checkpoint at the end of the experiment regardless of "
        "the checkpoint_freq.",
    )
    @click.option(
        "--custom-loggers/--no-custom-loggers",
        default=True,
        help="Use custom loggers from raylab.logger.",
    )
    @click.option(
        "--tune-log-level",
        type=str,
        default="WARN",
        show_default=True,
        help="Logging level for the trial executor process. This is independent from "
        "each trainer's logging level.",
    )
    @click.option(
        "--restore",
        type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
        default=None,
        show_default=True,
        help="Path to checkpoint. Only makes sense to set if running 1 trial.",
    )
    @functools.wraps(func)
    def wrapped(
        *args,
        name,
        local_dir,
        num_samples,
        stop,
        checkpoint_freq,
        checkpoint_at_end,
        custom_loggers,
        tune_log_level,
        restore,
        **kwargs,
    ):
        # pylint:disable=too-many-arguments
        from raylab.logger import DEFAULT_LOGGERS as CUSTOM_LOGGERS

        setup_tune_logger(tune_log_level)
        tune_kwargs = dict(
            name=name,
            local_dir=local_dir,
            num_samples=num_samples,
            checkpoint_freq=checkpoint_freq,
            checkpoint_at_end=checkpoint_at_end,
            restore=restore,
        )
        tune_kwargs["loggers"] = CUSTOM_LOGGERS if custom_loggers else None
        tune_kwargs["stop"] = dict(stop)

        return func(*args, tune_kwargs=tune_kwargs, **kwargs)

    return wrapped


def process_tune_kwargs(ctx, local_dir, name, **_):
    """Check missing/existing directories and prompt user if necessary."""
    create_if_necessary(ctx, local_dir)
    delete_if_necessary(ctx, osp.join(local_dir, name))


def setup_tune_logger(tune_log_level):
    """Set Tune's log level."""
    logging.getLogger("ray.tune").setLevel(tune_log_level)


def create_if_necessary(ctx, directory):
    """Ask user to allow directory creation if necessary."""
    msg = f"Directory {directory} does not exist. Create it?"
    if not osp.exists(directory):
        if not click.confirm(msg):
            ctx.exit()
        os.makedirs(directory)
        click.echo(f"Created directory {directory}")


def delete_if_necessary(ctx, exp_dir):
    """Ask user to allow directory deletion if necessary."""
    if osp.exists(exp_dir):
        msg = f"Experiment directory {exp_dir} already exists. Remove and proceed?"
        if not click.confirm(msg):
            ctx.exit()

        import shutil

        shutil.rmtree(exp_dir)
        click.echo(f"Removed directory {exp_dir}")



"""Collection of Exploration classes for `TorchPolicy`s."""
from .gaussian_noise import GaussianNoise
from .parameter_noise import ParameterNoise
from .random_uniform import RandomUniform
from .stochastic_actor import StochasticActor

__all__ = [
    "GaussianNoise",
    "ParameterNoise",
    "RandomUniform",
    "StochasticActor",
]



"""Policy for MAGE using PyTorch."""
from typing import List, Tuple

from nnrl.nn.critic import HardValue
from nnrl.optim import build_optimizer

from raylab.agents.sop import SOPTorchPolicy
from raylab.options import configure, option
from raylab.policy import EnvFnMixin
from raylab.policy.action_dist import WrapDeterministicPolicy
from raylab.policy.losses import MAGE, MaximumLikelihood
from raylab.policy.model_based.lightning import LightningModelTrainer, TrainingSpec
from raylab.policy.model_based.policy import MBPolicyMixin
from raylab.utils.types import StatDict


def default_model_training() -> dict:
    """Model training routine used by MAGE paper."""
    spec = TrainingSpec()
    spec.datamodule.holdout_ratio = 0.0
    spec.datamodule.max_holdout = 0
    spec.datamodule.batch_size = 256
    spec.datamodule.shuffle = True
    spec.datamodule.num_workers = 0
    spec.training.max_epochs = None
    spec.training.max_steps = 120
    spec.training.patience = None
    spec.warmup = spec.training
    return spec.to_dict()


@configure
@option("model_training", default=default_model_training())
@option("model_update_interval", default=25)
@option("improvement_steps", default=10, override=True)
@option("policy_delay", 2, override=True)
@option("batch_size", default=1024, override=True)
@option("lambda", default=0.05, help="TD error regularization for MAGE loss")
@option("module/type", "MAGE", override=True)
@option("optimizer/models", default={"type": "Adam", "lr": 1e-4, "weight_decay": 1e-4})
@option("optimizer/actor", default={"type": "Adam", "lr": 1e-4}, override=True)
@option("optimizer/critics", default={"type": "Adam", "lr": 1e-4}, override=True)
@option("exploration_config/pure_exploration_steps", 1000, override=True)
class MAGETorchPolicy(MBPolicyMixin, EnvFnMixin, SOPTorchPolicy):
    """MAGE policy in PyTorch to use with RLlib.

    Attributes:
        loss_model: maximum likelihood loss for model ensemble
        loss_actor: deterministic policy gradient loss
        loss_critic: model-based action-value-gradient estimator loss
    """

    # pylint:disable=too-many-ancestors
    dist_class = WrapDeterministicPolicy
    model_trainer: LightningModelTrainer

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)
        self._set_model_loss()
        self._set_critic_loss()
        self.build_timers()
        self.model_trainer = LightningModelTrainer(
            models=self.module.models,
            loss_fn=self.loss_model,
            optimizer=self.optimizers["models"],
            replay=self.replay,
            config=self.config,
        )

    def _set_model_loss(self):
        self.loss_model = MaximumLikelihood(self.module.models)

    def _set_critic_loss(self):
        module = self.module
        target_critic = HardValue(
            policy=module.target_actor, q_value=module.target_critics
        )
        self.loss_critic = MAGE(
            critics=module.critics,
            policy=module.actor,
            target_critic=target_critic,
            models=module.models,
        )
        self.loss_critic.gamma = self.config["gamma"]
        self.loss_critic.lambd = self.config["lambda"]

    def train_dynamics_model(
        self, warmup: bool = False
    ) -> Tuple[List[float], StatDict]:
        return self.model_trainer.optimize(warmup=warmup)

    def compile(self):
        super().compile()
        for loss in (self.loss_model, self.loss_actor, self.loss_critic):
            loss.compile()

    def _set_reward_hook(self):
        self.loss_critic.set_reward_fn(self.reward_fn)

    def _set_termination_hook(self):
        self.loss_critic.set_termination_fn(self.termination_fn)

    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["optimizer"]
        optimizers["models"] = build_optimizer(self.module.models, config["models"])
        return optimizers



"""Trainer and configuration for ACKTR."""
from raylab.agents.trpo import TRPOTrainer

from .policy import ACKTRTorchPolicy


class ACKTRTrainer(TRPOTrainer):
    """Single agent trainer for ACKTR."""

    _name = "ACKTR"
    _policy_class = ACKTRTorchPolicy



# pylint:disable=missing-module-docstring
from typing import Optional

import gym
import numpy as np
from gym.spaces import Box

from .mixins import IrrelevantRedundantMixin
from .utils import assert_flat_box_space, check_redundant_size_compat


class NonlinearRedundant(IrrelevantRedundantMixin, gym.ObservationWrapper):
    r"""Adds nonlinear redundant variables to the environment's observations.

    Computes redundant variables as nonlinear functions.
    .. math::
        (x_t, \cos(x_t), \sin(x_t)) \in \mathbb{R}^{3d}

    Args:
        env: Gym environment instance
        size: Number of left-most features from the observation to use in
            computing redundant variables. Defaults to the observation size
    """

    def __init__(self, env: gym.Env, size: Optional[int] = None):
        assert_flat_box_space(env.observation_space, self)
        super().__init__(env)
        original = self.env.observation_space
        self._size = size = size or original.shape[0]
        check_redundant_size_compat(size, original)

        low = np.concatenate([original.low, [-1] * 2 * size]).astype(original.dtype)
        high = np.concatenate([original.high, [1] * 2 * size]).astype(original.dtype)
        self.observation_space = Box(low=low, high=high)

        self._set_reward_if_possible()
        self._set_termination_if_possible()

    @property
    def added_size(self):
        return 2 * self.env.observation_space.shape[0]

    def _added_vars(self, observation: np.ndarray) -> np.ndarray:
        cos = np.cos(observation[: self._size])
        sin = np.sin(observation[: self._size])
        return np.concatenate([cos, sin])

    @staticmethod
    def wrap_env_function(func: callable, size: int) -> callable:
        return IrrelevantRedundantMixin.wrap_env_function(func, size * 2)



"""Modularized Q-Learning procedures."""
from abc import ABC, abstractmethod
from typing import List, Tuple

import torch
from nnrl.nn.critic import QValueEnsemble, VValue
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from torch import Tensor, nn

import raylab.utils.dictionaries as dutil
from raylab.utils.types import StatDict

from .abstract import Loss


class QLearningMixin(ABC):
    """Adds default call for Q-Learning losses."""

    # pylint:disable=too-few-public-methods
    batch_keys = (
        SampleBatch.CUR_OBS,
        SampleBatch.ACTIONS,
        SampleBatch.REWARDS,
        SampleBatch.NEXT_OBS,
        SampleBatch.DONES,
    )
    critics: QValueEnsemble

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, TensorDict]:
        """Compute loss for Q-value function."""
        obs, actions, rewards, next_obs, dones = dutil.get_keys(batch, *self.batch_keys)
        with torch.no_grad():
            target_values = self.critic_targets(rewards, next_obs, dones)
        loss_fn = nn.MSELoss()
        values = self.critics(obs, actions)
        critic_loss = torch.stack([loss_fn(v, target_values) for v in values]).sum()

        stats = {"loss(critics)": critic_loss.item()}
        stats.update(self.q_value_info(values))
        return critic_loss, stats

    @abstractmethod
    def critic_targets(
        self, rewards: Tensor, next_obs: Tensor, dones: Tensor
    ) -> Tensor:
        """Compute clipped 1-step approximation of Q^{\\pi}(s, a)."""

    @staticmethod
    def q_value_info(values: List[Tensor]) -> StatDict:
        """Return the average, min, and max Q-values in a batch."""
        info = {}
        # pylint:disable=invalid-name
        for i, q in enumerate(values):
            infoi = {
                f"Q{i}_mean": q.mean().item(),
                f"Q{i}_std": q.std().item(),
                f"Q{i}_max": q.max().item(),
                f"Q{i}_min": q.min().item(),
            }
            info.update(infoi)
        return info


class FittedQLearning(QLearningMixin, Loss):
    """Fitted Q-Learning.

    Generalized fitted Q-Learning implementation using a state-value function as
    target.

    Clipped Double Q-Learning can be achieved by composing the two target
    critics into a state-value function using, e.g., ClippedQValue and HardValue

    Args:
        critics: Main action-values
        target_critic: Target state-value function.

    Attributes:
        gamma: discount factor
    """

    gamma: float = 0.99

    def __init__(
        self,
        critics: QValueEnsemble,
        target_critic: VValue,
    ):
        self.critics = critics
        assert isinstance(
            target_critic, VValue
        ), "Need state-value function for critic target."
        self.target_critic = target_critic

    def critic_targets(self, rewards, next_obs, dones):
        values = self.target_critic(next_obs)
        values = torch.where(dones, torch.zeros_like(values), values)
        return rewards + self.gamma * values



"""Environment model handling mixins for TorchPolicy."""
from dataclasses import dataclass, field
from typing import List, Tuple

import numpy as np
import torch
from dataclasses_json import DataClassJsonMixin
from numpy.random import Generator
from ray.rllib import SampleBatch
from ray.rllib.utils import PiecewiseSchedule
from torch.nn import Module


@dataclass(frozen=True)
class SamplingSpec(DataClassJsonMixin):
    """Specifications for sampling from the model.

    Attributes:
        num_elites: Use this number of best performing models to sample
            transitions
        rollout_schedule: A list of tuples `(endpoint, value)`. The rollout
            length for timestep `t` is a linear interpolation between the two
            values corresponding to the nearest endpoints. Must be passed in
            increasing order of endpoints.
    """

    num_elites: int = 1
    rollout_schedule: List[Tuple[int, float]] = field(default_factory=lambda: [(0, 1)])

    def __post_init__(self):
        assert self.num_elites > 0, "Must have at least one elite model to sample from"
        assert all(
            a[0] <= b[0]
            for a, b in zip(self.rollout_schedule[:-1], self.rollout_schedule[1:])
        ), "Rollout schedule endpoints must be in increasing order"


class ModelSamplingMixin:
    """Adds model sampling behavior to a TorchPolicy class.

    Expects:
    * A `self.reward_fn` callable that computes the reward tensors for a batch
      of transitions
    * A `self.termination_fn` callable that computes the termination tensors for
      a batch of transitions
    * A `models` attribute in `self.module`
    * A `self.config` dict attribute
    * A `model_sampling` dict in `self.config`
    * A `seed` int in `self.config`

    Attributes:
        model_sampling_spec: Specifications for model training and sampling
        elite_models: Sequence of the `num_elites` best models sorted by
            performance. Initially set using the policy's model order.
        rng: Random number generator for choosing from the elite models for
            sampling.
    """

    model_sampling_spec: SamplingSpec
    rollout_schedule: PiecewiseSchedule
    elite_models: List[Module]
    rng: Generator

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_sampling_spec = spec = SamplingSpec.from_dict(
            self.config["model_sampling"]
        )

        self.rollout_schedule = PiecewiseSchedule(
            spec.rollout_schedule,
            framework="torch",
            outside_value=spec.rollout_schedule[-1][-1],
        )

        models = self.module.models
        num_elites = self.model_sampling_spec.num_elites
        assert num_elites <= len(models), "Cannot have more elites than models"
        self.elite_models = list(models[:num_elites])

        self.rng = np.random.default_rng(self.config["seed"])

    def set_new_elite(self, losses: List[float]):
        """Update the elite models based on model losses.

        Args:
            losses: list of model losses following the order of the ensemble
        """
        models = self.module.models
        self.elite_models = [models[i] for i in np.argsort(losses)]

    @torch.no_grad()
    def generate_virtual_sample_batch(self, samples: SampleBatch) -> SampleBatch:
        """Rollout model with latest policy.

        Produces samples for populating the virtual buffer, hence no gradient
        information is retained.

        If a transition is terminal, the next transition, if any, is generated from
        the initial state passed through `samples`.

        Args:
            samples: the transitions to extract initial states from

        Returns:
            A batch of transitions sampled from the model
        """
        virtual_samples = []
        obs = init_obs = self.convert_to_tensor(samples[SampleBatch.CUR_OBS])

        rollout_length = round(self.rollout_schedule(self.global_timestep))
        for _ in range(rollout_length):
            model = self.rng.choice(self.elite_models)

            action, _ = self.module.actor.sample(obs)
            next_obs, _ = model.sample(model(obs, action))
            reward = self.reward_fn(obs, action, next_obs)
            done = self.termination_fn(obs, action, next_obs)

            transition = {
                SampleBatch.CUR_OBS: obs,
                SampleBatch.ACTIONS: action,
                SampleBatch.NEXT_OBS: next_obs,
                SampleBatch.REWARDS: reward,
                SampleBatch.DONES: done,
            }
            virtual_samples += [
                SampleBatch({k: v.cpu().numpy() for k, v in transition.items()})
            ]
            obs = torch.where(done.unsqueeze(-1), init_obs, next_obs)

        return SampleBatch.concat_samples(virtual_samples)

    @staticmethod
    def model_sampling_defaults():
        """The default configuration dict for model sampling."""
        return SamplingSpec().to_dict()



"""Trainer and configuration for SVG(1) with maximum entropy."""
from raylab.agents import Trainer
from raylab.agents.model_based import set_policy_with_env_fn
from raylab.agents.off_policy import OffPolicyMixin
from raylab.options import configure

from .policy import SoftSVGTorchPolicy


@configure
@OffPolicyMixin.add_options
class SoftSVGTrainer(OffPolicyMixin, Trainer):
    """Single agent trainer for SoftSVG."""

    _name = "SoftSVG"
    _policy_class = SoftSVGTorchPolicy

    def after_init(self):
        super().after_init()
        set_policy_with_env_fn(self.workers, fn_type="reward")



"""NN architecture used in Deep Deterministic Policy Gradients."""
from dataclasses import dataclass, field

from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box
from nnrl.nn.actor.deterministic import DeterministicActor
from nnrl.nn.critic import ActionValueCritic
from torch import nn

ActorSpec = DeterministicActor.spec_cls
CriticSpec = ActionValueCritic.spec_cls


@dataclass
class DDPGSpec(DataClassJsonMixin):
    """Specifications for DDPG modules.

    Args:
        actor: Specifications for policy, behavior, and target policy
        critic: Specifications for action-value estimators
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments. Overrides actor and critic initializer
            specifications.
    """

    actor: ActorSpec = field(default_factory=ActorSpec)
    critic: CriticSpec = field(default_factory=CriticSpec)
    initializer: dict = field(default_factory=dict)

    def __post_init__(self):
        # Top-level initializer options take precedence over individual
        # component's options
        if self.initializer:
            self.actor.initializer = self.initializer
            self.critic.initializer = self.initializer


class DDPG(nn.Module):
    """NN module for DDPG-like algorithms.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for DDPG modules

    Attributes:
        actor (DeterministicPolicy): The deterministic policy to be learned
        behavior (DeterministicPolicy): The policy for exploration
        target_actor (DeterministicPolicy): The policy used for estimating the
            arg max in Q-Learning
        critics (QValueEnsemble): The action-value estimators to be learned
        target_critics (QValueEnsemble): The action-value estimators used for
            bootstrapping in Q-Learning
        spec_cls: Expected class of `spec` init argument
    """

    # pylint:disable=abstract-method
    spec_cls = DDPGSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: DDPGSpec):
        super().__init__()
        # Build actor
        actor = DeterministicActor(obs_space, action_space, spec.actor)
        self.actor = actor.policy
        self.behavior = actor.behavior
        self.target_actor = actor.target_policy

        # Build critic
        critic = ActionValueCritic(obs_space, action_space, spec.critic)
        self.critics = critic.q_values
        self.target_critics = critic.target_q_values



"""Loss functions for Maximum Likelihood Estimation."""
from typing import List, Tuple, Union

import torch
from nnrl.nn.model import SME, ForkedSME, StochasticModel
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from torch import Tensor, nn
from torch.jit import fork, wait

from raylab.utils.dictionaries import get_keys
from raylab.utils.types import StatDict

from .abstract import Loss


class LogVarReg(nn.Module):
    """Compute logvar bound penalty if needed."""

    # pylint:disable=abstract-method
    def forward(self, params: TensorDict) -> Tensor:
        """Compute logvar bound penalty if needed.

        Args:
            params: Model distribution parameters

        Returns:
            Regularization factors for model loss
        """
        # pylint:disable=no-self-use
        should_regularize = (
            "max_logvar" in params
            and params["max_logvar"].requires_grad
            and "min_logvar" in params
            and params["min_logvar"].requires_grad
        )
        if should_regularize:
            return 0.01 * params["max_logvar"].sum() - 0.01 * params["min_logvar"].sum()

        return torch.zeros(1).squeeze()


class NLLLoss(nn.Module):
    """Compute Negative Log-Likelihood loss."""

    # pylint:disable=abstract-method

    def __init__(self, model: StochasticModel):
        super().__init__()
        self.model = model
        self.logvar_reg = LogVarReg()

    def forward(self, obs: Tensor, act: Tensor, new_obs: Tensor) -> Tensor:
        # pylint:disable=missing-function-docstring
        params = self.model(obs, act)
        logp = self.model.log_prob(new_obs, params)
        nll = -logp.mean()
        regularizer = self.logvar_reg(params)
        return nll + regularizer


class Losses(nn.ModuleList):
    # pylint:disable=abstract-method,missing-class-docstring
    def __init__(self, losses: List[NLLLoss]):
        assert all(isinstance(loss, NLLLoss) for loss in losses)
        super().__init__(losses)

    def forward(self, obs: Tensor, act: Tensor, new_obs: Tensor) -> List[Tensor]:
        # pylint:disable=arguments-differ
        return [loss(obs, act, new_obs) for loss in self]


class ForkedLosses(Losses):
    # pylint:disable=abstract-method,missing-class-docstring
    def forward(self, obs: Tensor, act: Tensor, new_obs: Tensor) -> List[Tensor]:
        futures = [fork(loss, obs, act, new_obs) for loss in self]
        return [wait(f) for f in futures]


class MaximumLikelihood(Loss):
    """Loss function for model learning of single transitions.

    Args:
        models: parametric stochastic model (single or ensemble)
    """

    batch_keys: Tuple[str, str, str] = (
        SampleBatch.CUR_OBS,
        SampleBatch.ACTIONS,
        SampleBatch.NEXT_OBS,
    )
    _last_output: Tuple[Tensor, StatDict]

    def __init__(self, models: Union[StochasticModel, SME]):
        if isinstance(models, StochasticModel):
            # Treat everything as if ensemble
            models = SME([models])
        self.models = models
        self.tag = "nll"
        self.build_losses()
        self._last_losses = torch.zeros(len(self.models))

    def build_losses(self):
        # pylint:disable=missing-function-docstring
        models = self.models
        losses = [NLLLoss(m) for m in models]
        cls = ForkedLosses if isinstance(models, ForkedSME) else Losses
        self.loss_fns = cls(losses)

    @property
    def last_output(self) -> Tuple[Tensor, StatDict]:
        """Last computed losses for each individual model and associated info."""
        return self._last_output

    def compile(self):
        # pylint:disable=missing-function-docstring,attribute-defined-outside-init
        self.loss_fns = torch.jit.script(self.loss_fns)

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        """Compute Maximum Likelihood Estimation (MLE) model loss.

        Returns:
            A tuple with a 1d loss tensor containing each model's loss and a
            dictionary of loss statistics
        """
        obs, act, new_obs = get_keys(batch, *self.batch_keys)
        nlls = self.loss_fns(obs, act, new_obs)

        losses = torch.stack(nlls)
        info = {f"{self.tag}(models[{i}])": n.item() for i, n in enumerate(nlls)}
        self._last_output = (losses, info)
        return losses.mean(), info



"""Custom loggers to use with Tune."""
from ray.tune.logger import CSVLogger

from .progress_json_logger import ProgressJsonLogger

DEFAULT_LOGGERS = (ProgressJsonLogger, CSVLogger)

try:
    from .torch_tensorboard_logger import TorchTBLogger

    DEFAULT_LOGGERS = DEFAULT_LOGGERS + (TorchTBLogger,)
except ImportError:
    pass



"""Mixins for model-based policies."""
from .envfn import EnvFnMixin
from .lightning import LightningModelTrainer
from .policy import MBPolicyMixin
from .sampling import ModelSamplingMixin

__all__ = [
    "EnvFnMixin",
    "LightningModelTrainer",
    "MBPolicyMixin",
    "ModelSamplingMixin",
]



# pylint:disable=missing-module-docstring
import nnrl.utils as ptu
import numpy as np

from .base import BaseExploration, Model


class RandomUniform(BaseExploration):
    """Samples actions from the Gym action space

    Args:
        pure_exploration_steps (int): Number of initial timesteps to explore.
            Until this many timesteps have elapsed, the agent's policy will be
            ignored & it will instead take uniform random actions. Can be used in
            conjunction with learning_starts (which controls when the first
            optimization step happens) to decrease dependence of exploration &
            optimization on initial policy parameters.
    """

    # pylint:disable=too-few-public-methods

    def __init__(self, *args, pure_exploration_steps=0, **kwargs):
        super().__init__(*args, **kwargs)
        if self.framework != "torch":
            raise ValueError(
                f"{type(self)} incompatible with '{self.framework}' framework."
            )
        self._pure_exploration_steps = pure_exploration_steps

    def get_exploration_action(self, *, action_distribution, timestep, explore=True):
        # pylint:disable=unused-argument
        if explore:
            obs = action_distribution.inputs["obs"]
            acts = ptu.convert_to_tensor(
                [self.action_space.sample() for _ in range(obs.size(0))], obs.device
            )
            logp = ptu.convert_to_tensor(
                [-np.log(self.action_space.high - self.action_space.low).sum(axis=-1)]
                * obs.size(0),
                obs.device,
            )
            return acts, logp
        return action_distribution.deterministic_sample()

    @classmethod
    def check_model_compat(cls, model: Model):
        pass



# pylint:disable=missing-module-docstring
from typing import Callable, Iterable

from ray.rllib.evaluation.worker_set import WorkerSet
from ray.rllib.execution.metric_ops import StandardMetricsReporting
from ray.rllib.execution.rollout_ops import ParallelRollouts
from ray.rllib.execution.train_ops import TrainOneStep
from ray.rllib.utils.typing import ResultDict, TrainerConfigDict

from raylab.execution import LearningStarts
from raylab.options import option


def off_policy_execution_plan(workers: WorkerSet, config: TrainerConfigDict):
    """RLlib's default execution plan with an added warmup phase."""
    # Collects experiences in parallel from multiple RolloutWorker actors.
    rollouts = ParallelRollouts(workers, mode="bulk_sync")
    # On the first iteration, combine experience batches until we hit `learning_starts`
    # in size.
    rollouts = rollouts.combine(
        LearningStarts(learning_starts=config["learning_starts"])
    )
    # Then, train the policy on those experiences and update the workers.
    train_op = rollouts.for_each(TrainOneStep(workers))

    # Add on the standard episode reward, etc. metrics reporting. This returns
    # a LocalIterator[metrics_dict] representing metrics for each train step.
    return StandardMetricsReporting(train_op, workers, config)


class OffPolicyMixin:
    """Mixin for off-policy agents."""

    # pylint:disable=missing-function-docstring
    def validate_config(self, config: dict):
        super().validate_config(config)
        assert config["num_workers"] == 0, "No point in using additional workers."
        assert (
            config["rollout_fragment_length"] >= 1
        ), "At least one sample must be collected."

    @property
    def execution_plan(
        self,
    ) -> Callable[[WorkerSet, TrainerConfigDict], Iterable[ResultDict]]:
        return off_policy_execution_plan

    @staticmethod
    def add_options(trainer_cls: type) -> type:
        cls = trainer_cls
        for opt in [
            option(
                "learning_starts",
                default=0,
                help="Hold this number of timesteps before first training operation.",
            ),
            option("rollout_fragment_length", default=1, override=True),
            option("num_workers", default=0, override=True),
        ]:
            cls = opt(cls)
        return cls



"""CLI utilities for RayLab."""
from __future__ import annotations

import click

from .best_checkpoint import find_best
from .evaluate_checkpoint import rollout
from .experiment import experiment
from .info import info_cli


@click.group()
def raylab():
    """RayLab: Reinforcement learning algorithms in RLlib."""


@raylab.command()
@click.argument(
    "paths",
    nargs=-1,
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=False),
)
def dashboard(paths: tuple[str, ...]):
    """Launch the experiment dashboard to monitor training progress."""
    import subprocess

    from . import experiment_dashboard

    subprocess.run(
        ["streamlit", "run", experiment_dashboard.__file__] + list(paths), check=True
    )


@raylab.command()
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
)
def episodes(path: str):
    """Launch the episode dashboard to monitor state and action distributions."""
    import subprocess

    from . import episode_dashboard

    subprocess.run(["streamlit", "run", episode_dashboard.__file__, path], check=True)


@raylab.command()
@click.argument("agent_id", type=str)
@click.argument(
    "checkpoint",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
)
def test_module(agent_id, checkpoint):
    """Launch dashboard to test generative models from a checkpoint."""
    import subprocess

    from . import module_dashboard

    subprocess.run(
        ["streamlit", "run", module_dashboard.__file__, agent_id, checkpoint],
        check=True,
    )


raylab.add_command(experiment)
raylab.add_command(find_best)
raylab.add_command(rollout)
raylab.add_command(info_cli)



"""Episode monitoring with Streamlit."""
import os.path as osp
import shelve

import numpy as np
import pandas as pd
import streamlit as st
from bokeh.layouts import gridplot
from bokeh.plotting import figure

# pylint:disable=invalid-name,missing-docstring,pointless-string-statement
# pylint:disable=no-value-for-parameter
"""
# Visualize Episode
"""


def main():
    # pylint:disable=too-many-locals
    import sys

    assert len(sys.argv) <= 2, "Only one episode log from `rllib rollout` allowed."
    path, _ = osp.splitext(sys.argv[1])

    episodes = []
    with shelve.open(path) as rollouts:
        for episode_index in range(rollouts["num_episodes"]):
            episodes += [rollouts[str(episode_index)]]

    rows = []
    for idx, episode in enumerate(episodes):
        for timestep, transition in enumerate(episode):
            obs, act, _, rew, done, *_ = transition
            row = {"episode": idx, "timestep": timestep, "reward": rew, "done": done}
            for dim, ob_ in enumerate(obs):
                row[f"obs[{dim}]"] = ob_
            for dim, ac_ in enumerate(act):
                row[f"act[{dim}]"] = ac_
            rows += [row]

    keys = rows[0].keys()
    data = pd.DataFrame(data={k: np.stack([r[k] for r in rows]) for k in keys})
    data["done"] = data["done"].astype(np.float32)

    stats_columns = [c for c in data.columns if c not in {"timestep", "episode"}]
    grouped = data.groupby("timestep")[stats_columns]
    means = grouped.mean()
    stds = grouped.std()

    pics = []
    for key in stats_columns:
        pic = figure(title=key)
        pic.xaxis.axis_label = "timestep"
        pic.yaxis.axis_label = "value"

        # pylint:disable=too-many-function-args
        pic.line(means.index, means[key])
        pic.varea(
            means.index,
            y1=means[key] - stds[key],
            y2=means[key] + stds[key],
            fill_alpha=0.25,
        )
        # pylint:enable=too-many-function-args

        pics += [[pic]]

    st.bokeh_chart(
        gridplot(
            pics,
            plot_width=350,
            plot_height=100,
            sizing_mode="scale_both",
            toolbar_location="right",
        )
    )


if __name__ == "__main__":
    main()



"""Mixins for Gym environment wrappers."""
from abc import ABC, abstractmethod
from typing import List, Optional

import numpy as np
from gym.utils import seeding

from .utils import ignore_rightmost_variables


class RNGMixin:
    """Adds a separate random number generator to an environment wrapper.

    Appends the wrapper's rng seed to the list of seeds returned by
    :meth:`seed`.

    Attributes:
        np_random: A numpy RandomState
    """

    # pylint:disable=missing-function-docstring,too-few-public-methods
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.np_random, _ = seeding.np_random()

    def seed(self, seed: Optional[int] = None) -> List[int]:
        seeds = super().seed(seed) or []
        self.np_random, seed_ = seeding.np_random(seed)
        return seeds + [seed_]


class IrrelevantRedundantMixin(ABC):
    """Common irrelevant/redundant observation wrapper interface."""

    @property
    @abstractmethod
    def added_size(self):
        """Number of right-most irrelevant/redundant state dimensions."""

    def observation(self, observation: np.ndarray) -> np.ndarray:
        """Concatenate irrelevant/redundant variables to the observation."""
        irrelevant_or_redundant = self._added_vars(observation)
        observation = np.concatenate([observation, irrelevant_or_redundant])
        return observation.astype(self.observation_space.dtype)

    @staticmethod
    def wrap_env_function(func: callable, size: int) -> callable:
        """Wrap base env reward/termination function to ignore added variables.

        Args:
            func: Callable for reward/termination function
            size: Number of irrelevant/redundant variables
        """
        return ignore_rightmost_variables(func, size)

    @abstractmethod
    def _added_vars(self, observation: np.ndarray) -> np.ndarray:
        pass

    def _set_reward_if_possible(self):
        if hasattr(self.env, "reward_fn"):
            self.reward_fn = ignore_rightmost_variables(
                self.env.reward_fn, self.added_size
            )

    def _set_termination_if_possible(self):
        if hasattr(self.env, "termination_fn"):
            self.termination_fn = ignore_rightmost_variables(
                self.env.termination_fn, self.added_size
            )



"""NN architecture used in Deep Deterministic Policy Gradients."""
from dataclasses import dataclass, field

from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box
from nnrl.nn.actor.deterministic import DeterministicActor
from nnrl.nn.critic import ActionValueCritic
from torch import nn


@dataclass
class ActorSpec(DeterministicActor.spec_cls):
    """Specifications for policy, behavior, and target policy."""

    separate_target_policy: bool = field(default=True, init=False, repr=False)
    smooth_target_policy: bool = field(default=True, init=False, repr=False)


CriticSpec = ActionValueCritic.spec_cls


@dataclass
class TD3Spec(DataClassJsonMixin):
    """Specifications for TD3 modules.

    Args:
        actor: Specifications for policy, behavior, and target policy
        critic: Specifications for action-value estimators
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments. Overrides actor and critic initializer
            specifications.
    """

    actor: ActorSpec = field(default_factory=ActorSpec)
    critic: CriticSpec = field(default_factory=CriticSpec)
    initializer: dict = field(default_factory=dict)

    def __post_init__(self):
        # Top-level initializer options take precedence over individual
        # component's options
        if self.initializer:
            self.actor.initializer = self.initializer
            self.critic.initializer = self.initializer


class TD3(nn.Module):
    """NN module for TD3-like algorithms.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for DDPG modules

    Attributes:
        actor (DeterministicPolicy): The deterministic policy to be learned
        behavior (DeterministicPolicy): The policy for exploration
        target_actor (DeterministicPolicy): The policy used for estimating the
            arg max in Q-Learning
        critics (QValueEnsemble): The action-value estimators to be learned
        target_critics (QValueEnsemble): The action-value estimators used for
            bootstrapping in Q-Learning
        spec_cls: Expected class of `spec` init argument
    """

    # pylint:disable=abstract-method
    spec_cls = TD3Spec

    def __init__(self, obs_space: Box, action_space: Box, spec: TD3Spec):
        super().__init__()
        # Build actor
        actor = DeterministicActor(obs_space, action_space, spec.actor)
        self.actor = actor.policy
        self.behavior = actor.behavior
        self.target_actor = actor.target_policy
        main, target = set(self.actor.parameters()), set(self.target_actor.parameters())
        assert not main.intersection(
            target
        ), "Main and target policy cannot share parameters"
        for par in target:
            par.requires_grad = False

        # Build critic
        critic = ActionValueCritic(obs_space, action_space, spec.critic)
        self.critics = critic.q_values
        self.target_critics = critic.target_q_values



# pylint:disable=missing-module-docstring
import gym
import numpy as np
import torch
from torch import Tensor

DEFAULT_CONFIG = {
    "start": [0.0, 1.0],
    "end": [8.0, 9.0],
    "action_lower_bound": [-1.0, -1.0],
    "action_upper_bound": [1.0, 1.0],
    "deceleration_zones": {"center": [[0.0, 0.0]], "decay": [2.0]},
    "noise": {"loc": [0.0, 0.0], "scale_tril": [[0.3, 0.0], [0.0, 0.3]]},
    "horizon": 20,
    "init_dist": True,
}


class NavigationEnv(gym.Env):
    """NavigationEnv implements a gym environment for the Navigation
    domain.

    The agent must navigate from a start position to and end position.
    Its actions represent displacements in the 2D plane. Gaussian noise
    is added to the final position as to incorporate uncertainty in the
    transition. Additionally, the effect of an action might be decreased
    by a scalar factor dependent on the proximity of deceleration zones.

    Please refer to the AAAI paper for further details:

    Bueno, T.P., de Barros, L.N., Mauá, D.D. and Sanner, S., 2019, July.
    Deep Reactive Policies for Planning in Stochastic Nonlinear Domains.
    In Proceedings of the AAAI Conference on Artificial Intelligence.
    """

    # pylint:disable=too-many-instance-attributes

    metadata = {"render.modes": ["human"]}

    def __init__(self, config=None):
        self._config = {**DEFAULT_CONFIG, **(config or {})}

        self._start = np.array(self._config["start"], dtype=np.float32)
        self._end = np.array(self._config["end"], dtype=np.float32)

        self._min_x = self._start[0] - 1.0
        self._max_x = self._end[0] + 1.0
        self._min_y = self._start[1] - 1.0
        self._max_y = self._end[1] + 1.0

        self.observation_space = gym.spaces.Box(
            low=np.array([-np.inf, -np.inf, 0.0], dtype=np.float32),
            high=np.array([np.inf, np.inf, 1.0], dtype=np.float32),
        )
        self.action_space = gym.spaces.Box(
            low=np.array(self._config["action_lower_bound"], dtype=np.float32),
            high=np.array(self._config["action_upper_bound"], dtype=np.float32),
        )

        self._deceleration_zones = self._config["deceleration_zones"]
        if self._deceleration_zones:
            self._deceleration_decay = np.array(
                self._deceleration_zones["decay"], dtype=np.float32
            )
            self._deceleration_center = np.array(
                self._deceleration_zones["center"], dtype=np.float32
            )

        self._noise = self._config["noise"]
        self._horizon = self._config["horizon"]
        self._state = None

    def reset(self):
        if self._config["init_dist"]:
            xpos = np.random.uniform(self._min_x, self._max_x)
            ypos = np.random.uniform(self._min_y, self._max_y)
            time = 0.0
            self._state = np.array([xpos, ypos, time], dtype=np.float32)
        else:
            self._state = np.append(self._start, np.array(0.0, dtype=np.float32))

        return self._state

    @torch.no_grad()
    def step(self, action):
        state, action = map(torch.as_tensor, (self._state, action))
        next_state, _ = self.transition_fn(state, action)
        reward = self.reward_fn(state, action, next_state).item()
        self._state = next_state.numpy()
        return self._state, reward, self._terminal(), {}

    def transition_fn(self, state, action, sample_shape=()):
        # pylint:disable=missing-docstring
        position, time = self._unpack_state(state)
        deceleration = 1.0
        if self._deceleration_zones:
            deceleration = self._deceleration(position)

        position = position + (deceleration * action)
        position, logp = self._sample_noise(position, sample_shape)
        time = self._step_time(time)
        time = time.expand_as(position[..., -1:])
        return torch.cat([position, time], dim=-1), logp

    def _deceleration(self, position: Tensor) -> Tensor:
        # Decay is a tensor of shape (n)
        decay = torch.from_numpy(self._deceleration_decay)
        # Center is a matrix with rows as center positions (n, d)
        center = torch.from_numpy(self._deceleration_center)
        # Consider position as a batched row vector (*, d)
        # Unsqueeze position to form a (*, 1, d) matrix
        position = position.unsqueeze(-2)
        # Broadcasting will reshape position and center to (*, n, d)
        # Leverage broadcasting rules to compute batched distances
        # distance is a vector with distances to each center (*, n)
        distance = self.distance(position, center)
        # Compute decelerations for each zone (*, n)
        unreduced_decel = 2 / (1.0 + torch.exp(-decay * distance)) - 1.0
        # Reduce by computing the product of decelerations (*, 1)
        deceleration = torch.prod(unreduced_decel, dim=-1, keepdim=True)
        return deceleration

    def _sample_noise(self, position, sample_shape):
        loc = position + torch.as_tensor(self._noise["loc"])
        scale_tril = torch.as_tensor(self._noise["scale_tril"])
        dist = torch.distributions.MultivariateNormal(loc=loc, scale_tril=scale_tril)
        sample = dist.rsample(sample_shape=sample_shape)
        return sample, dist.log_prob(sample.detach())

    def _step_time(self, time):
        timestep = torch.round(self._horizon * time)
        return torch.clamp((timestep + 1) / self._horizon, 0, 1)

    def reward_fn(self, state: Tensor, action: Tensor, next_state: Tensor) -> Tensor:
        # pylint:disable=unused-argument,missing-docstring
        position, _ = self._unpack_state(next_state)
        goal = torch.from_numpy(self._end)
        return torch.norm(position - goal, dim=-1).neg()

    def _terminal(self):
        position, time = self._unpack_state(self._state)
        return np.allclose(position, self._end, atol=1e-1) or time.item() >= 1.0

    def termination_fn(
        self, state: Tensor, action: Tensor, next_state: Tensor
    ) -> Tensor:
        # pylint:disable=unused-argument,missing-docstring
        position, time = self._unpack_state(next_state)
        goal = torch.from_numpy(self._end)
        hit_goal = self.distance(position, goal) <= 1e-1
        timeout = time[..., 0] >= 1.0
        return hit_goal | timeout

    @staticmethod
    def distance(pos1: Tensor, pos2: Tensor) -> Tensor:
        """Batched distance computation between positions.

        Returns:
            Tensor with distances for each batch dimension
        """
        return torch.norm(pos1 - pos2, dim=-1)

    @staticmethod
    def _unpack_state(state):
        position, time = state[..., :-1], state[..., -1:]
        return position, time

    def render(self, mode="human"):
        pass



"""Trainer and default config for MAGE."""
from raylab.agents import Trainer
from raylab.agents.model_based import ModelBasedMixin
from raylab.agents.off_policy import OffPolicyMixin
from raylab.options import configure

from .policy import MAGETorchPolicy


@configure
@OffPolicyMixin.add_options
class MAGETrainer(ModelBasedMixin, OffPolicyMixin, Trainer):
    """Single agent trainer for MAGE."""

    _name = "MAGE"
    _policy_class = MAGETorchPolicy



"""ACKTR policy implemented in PyTorch."""
import warnings

import numpy as np
import torch
from nnrl.nn.distributions import Normal
from nnrl.optim import build_optimizer
from nnrl.optim.hessian_free import line_search
from nnrl.optim.kfac import KFACMixin
from ray.rllib import SampleBatch
from ray.rllib.evaluation.postprocessing import Postprocessing, compute_advantages
from ray.rllib.utils import override
from torch import nn

import raylab.utils.dictionaries as dutil
from raylab.agents.trpo.policy import LINESEARCH_DEFAULTS
from raylab.options import configure, option
from raylab.policy import TorchPolicy, learner_stats
from raylab.policy.action_dist import WrapStochasticPolicy
from raylab.utils.explained_variance import explained_variance

DEFAULT_OPTIM_CONFIG = {
    "actor": {
        # Arguments for KFAC
        "type": "KFAC",
        "eps": 1e-3,
        "sua": False,
        "pi": True,
        "update_freq": 1,
        "alpha": 0.95,
        "kl_clip": 1e-2,
        "eta": 1.0,
        "lr": 1.0,
    },
    "critic": {
        # Can choose different optimizer
        "type": "KFAC",
        "eps": 1e-3,
        "sua": False,
        "pi": True,
        "update_freq": 1,
        "alpha": 0.95,
        "kl_clip": 1e-2,
        "eta": 1.0,
        "lr": 1.0,
    },
}


@configure
@option(
    "fvp_samples",
    10,
    help="Number of actions to sample per state for Fisher vector-product calculation",
)
@option("lambda", 0.97, help=r"For GAE(\gamma, \lambda)")
@option("val_iters", 80, help="Number of iterations to fit value function")
@option("use_gae", True, help="Whether to use Generalized Advantage Estimation")
@option(
    "line_search",
    True,
    help="""
    Whether to use a line search to calculate policy update.
    Effectively turns TRPO into Natural PG when turned off.
    """,
)
@option("line_search_options", LINESEARCH_DEFAULTS)
@option("module/type", "TRPO")
@option("optimizer", DEFAULT_OPTIM_CONFIG, override=True)
@option("exploration_config/type", "raylab.utils.exploration.StochasticActor")
class ACKTRTorchPolicy(TorchPolicy):
    """Policy class for Actor-Critic with Kronecker factored Trust Region."""

    # pylint:disable=abstract-method
    dist_class = WrapStochasticPolicy

    @override(TorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = dutil.deep_merge(
            DEFAULT_OPTIM_CONFIG,
            self.config["optimizer"],
            new_keys_allowed=False,
            allow_new_subkey_list=[],
            override_all_if_type_changes=["actor", "critic"],
        )
        assert config["actor"]["type"] in [
            "KFAC",
            "EKFAC",
        ], "ACKTR must use optimizer with Kronecker Factored curvature estimation."

        mapping = {
            "actor": build_optimizer(self.module.actor, config["actor"]),
            "critic": build_optimizer(self.module.critic, config["critic"]),
        }

        optimizers.update(mapping)
        return optimizers

    @override(TorchPolicy)
    def compile(self):
        warnings.warn(f"{type(self).__name__} is incompatible with TorchScript")

    @torch.no_grad()
    @override(TorchPolicy)
    def postprocess_trajectory(
        self, sample_batch, other_agent_batches=None, episode=None
    ):
        sample_batch = super().postprocess_trajectory(
            sample_batch, other_agent_batches=other_agent_batches, episode=episode
        )

        last_obs = self.convert_to_tensor(sample_batch[SampleBatch.NEXT_OBS][-1])
        last_r = self.module.critic(last_obs).squeeze(-1).numpy()

        cur_obs = self.convert_to_tensor(sample_batch[SampleBatch.CUR_OBS])
        sample_batch[SampleBatch.VF_PREDS] = (
            self.module.critic(cur_obs).squeeze(-1).numpy()
        )
        sample_batch = compute_advantages(
            sample_batch,
            last_r,
            gamma=self.config["gamma"],
            lambda_=self.config["lambda"],
            use_gae=self.config["use_gae"],
        )
        return sample_batch

    @learner_stats
    @override(TorchPolicy)
    def learn_on_batch(self, samples):
        batch_tensors = self.lazy_tensor_dict(samples)
        info = {}

        info.update(self._update_actor(batch_tensors))
        info.update(self._update_critic(batch_tensors))
        info.update(self.extra_grad_info(batch_tensors))
        info.update(self.get_exploration_info())

        return info

    def _update_actor(self, batch_tensors):
        info = {}
        cur_obs, actions, advantages = dutil.get_keys(
            batch_tensors,
            SampleBatch.CUR_OBS,
            SampleBatch.ACTIONS,
            Postprocessing.ADVANTAGES,
        )
        advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)

        # Compute whitening matrices
        sshape = (self.config["fvp_samples"],)
        with self.optimizers["actor"].record_stats():
            _, log_prob = self.module.actor.sample(cur_obs.requires_grad_(), sshape)
            log_prob.mean().backward()

        # Compute surrogate loss
        with self.optimizers.optimize("actor"):
            surr_loss = -(
                self.module.actor.log_prob(cur_obs, actions) * advantages
            ).mean()
            info["loss(actor)"] = surr_loss.item()
            surr_loss.backward()
            pol_grad = [p.grad.clone() for p in self.module.actor.parameters()]

        if self.config["line_search"]:
            info.update(self._perform_line_search(pol_grad, surr_loss, batch_tensors))

        return info

    def _perform_line_search(self, pol_grad, surr_loss, batch_tensors):
        # pylint:disable=too-many-locals
        kl_clip = self.optimizers["actor"].state["kl_clip"]
        expected_improvement = sum(
            (g * p.grad.data).sum()
            for g, p in zip(pol_grad, self.module.actor.parameters())
        ).item()

        cur_obs, actions, old_logp, advantages = dutil.get_keys(
            batch_tensors,
            SampleBatch.CUR_OBS,
            SampleBatch.ACTIONS,
            SampleBatch.ACTION_LOGP,
            Postprocessing.ADVANTAGES,
        )

        @torch.no_grad()
        def f_barrier(scale):
            for par in self.module.actor.parameters():
                par.data.add_(par.grad.data, alpha=scale)
            new_logp = self.module.actor.log_prob(cur_obs, actions)
            for par in self.module.actor.parameters():
                par.data.sub_(par.grad.data, alpha=scale)
            surr_loss = self._compute_surr_loss(old_logp, new_logp, advantages)
            avg_kl = torch.mean(old_logp - new_logp)
            return surr_loss.item() if avg_kl < kl_clip else np.inf

        scale, expected_improvement, improvement = line_search(
            f_barrier,
            1,
            1,
            expected_improvement,
            y_0=surr_loss.item(),
            **self.config["line_search_options"],
        )
        improvement_ratio = (
            improvement / expected_improvement if expected_improvement else np.nan
        )
        info = {
            "expected_improvement": expected_improvement,
            "actual_improvement": improvement,
            "improvement_ratio": improvement_ratio,
        }
        for par in self.module.actor.parameters():
            par.data.add_(par.grad.data, alpha=scale)
        return info

    @staticmethod
    def _compute_surr_loss(old_logp, new_logp, advantages):
        return -torch.mean(torch.exp(new_logp - old_logp) * advantages)

    def _update_critic(self, batch_tensors):
        cur_obs, value_targets = dutil.get_keys(
            batch_tensors,
            SampleBatch.CUR_OBS,
            Postprocessing.VALUE_TARGETS,
        )
        mse = nn.MSELoss()
        fake_dist = Normal()
        fake_scale = torch.ones_like(value_targets)

        for _ in range(self.config["val_iters"]):
            if isinstance(self.optimizers["critic"], KFACMixin):
                # Compute whitening matrices
                with self.optimizers["critic"].record_stats():
                    values = self.module.critic(cur_obs).squeeze(-1)
                    fake_samples = values + torch.randn_like(values)
                    log_prob = fake_dist.log_prob(
                        fake_samples.detach(), {"loc": values, "scale": fake_scale}
                    )
                    log_prob.mean().backward()

            with self.optimizers.optimize("critic"):
                mse_loss = mse(self.module.critic(cur_obs).squeeze(-1), value_targets)
                mse_loss.backward()

        return {"loss(critic)": mse_loss.item()}

    @torch.no_grad()
    def extra_grad_info(self, batch_tensors):  # pylint:disable=unused-argument
        """Return statistics right after components are updated."""
        cur_obs, actions, old_logp, value_targets, value_preds = dutil.get_keys(
            batch_tensors,
            SampleBatch.CUR_OBS,
            SampleBatch.ACTIONS,
            SampleBatch.ACTION_LOGP,
            Postprocessing.VALUE_TARGETS,
            SampleBatch.VF_PREDS,
        )

        info = {
            "kl_divergence": torch.mean(
                old_logp - self.module.actor.log_prob(cur_obs, actions)
            ).item(),
            "entropy": torch.mean(-old_logp).item(),
            "perplexity": torch.mean(-old_logp).exp().item(),
            "explained_variance": explained_variance(
                value_targets.numpy(), value_preds.numpy()
            ),
        }
        info.update(
            {
                f"grad_norm({k})": nn.utils.clip_grad_norm_(
                    getattr(self.module, k).parameters(), float("inf")
                ).item()
                for k in ("actor", "critic")
            }
        )
        return info
